from nicegui import ui, app
import markdown2, re
from biblemateweb import get_translation
from agentmake.plugins.uba.lib.BibleParser import BibleVerseParser

class Notepad:
    def __init__(self, content):
        self.text_content = content
        self.is_editing = True
        self.parser = BibleVerseParser(False, language=app.storage.user['ui_language'])
        
    def setup_ui(self):

        # Dialog to confirm deleting a note
        with ui.dialog() as delete_dialog, ui.card():
            ui.label('Are you sure you want to delete this note?')
            with ui.row().classes('justify-end w-full'):
                ui.button('Cancel', on_click=delete_dialog.close).props('flat text-color=secondary')
                ui.button('Delete', color='red', on_click=lambda: (self.clear_text(), delete_dialog.close()))
        # --- Toolbar ---
        with ui.row().classes('gap-2 mb-0 w-full items-center'):
            self.mode_btn = ui.button(get_translation('Read Mode'), icon='visibility', on_click=self.toggle_mode).props('color=primary')
            ui.button(on_click=self.download_file, icon='download').props('flat round color=secondary').tooltip(get_translation("Download"))
            ui.button(on_click=lambda: self.upload.run_method('pickFiles'), icon='upload').props('flat round color=secondary').tooltip(get_translation("Import"))
            ui.button(on_click=delete_dialog.open, icon='delete').props('flat round color=negative').tooltip(get_translation("Clear"))
            self.upload = ui.upload(
                on_upload=self.handle_upload, 
                auto_upload=True
            ).props('accept=.txt,.md').classes('hidden')

        # --- Content Area ---
        # Card must be 'flex flex-col' so the child (textarea) can grow
        with ui.card().classes('w-full h-[70vh] p-0 flex flex-col'):
            # 1. Edit Mode: Text Area
            # We apply our custom 'full-height-textarea' class here
            self.textarea = ui.textarea(
                placeholder='Start typing your notes here...',
                value=self.text_content
            ).classes('w-full flex-grow full-height-textarea p-2 border-none focus:outline-none') \
             .props('flat squares resize-none') \
             .bind_visibility_from(self, 'is_editing')

            # 2. Read Mode: HTML Preview
            with ui.scroll_area().classes('w-full flex-grow p-2') \
                    .bind_visibility_from(self, 'is_editing', backward=lambda x: not x):
                self.html_view = ui.html(f'<div class="content-text">{self.text_content}</div>', sanitize=False).classes('w-full prose max-w-none')
        
        ui.label('💡 Tip: Switch to Read Mode to see the formatted result.').classes('text-sm text-gray-600 mt-4')

    def toggle_mode(self):
        self.is_editing = not self.is_editing
        if self.is_editing:
            self.mode_btn.text = 'Read Mode'
            self.mode_btn.props('icon=visibility')
        else:
            self.mode_btn.text = get_translation('Edit Mode')
            self.mode_btn.props('icon=edit')
            self.update_preview()

    def update_preview(self):
        content = self.textarea.value or ''
        try:
            # Added your requested extras
            html_content = markdown2.markdown(
                content, 
                extras=["tables", "fenced-code-blocks", "toc", "cuddled-lists"]
            )
            html_content = self.parser.parseText(html_content)
            html_content = re.sub(r'''(onclick|ondblclick)="(bdbid|lex|cr|bcv|website)\((.*?)\)"''', r'''\1="emitEvent('\2', [\3]); return false;"''', html_content)
            html_content = re.sub(r"""(onclick|ondblclick)='(bdbid|lex|cr|bcv|website)\((.*?)\)'""", r"""\1='emitEvent("\2", [\3]); return false;'""", html_content)
            self.html_view.content = html_content
        except Exception as e:
            self.html_view.content = f"<p class='text-red-500'>Error: {str(e)}</p>"

    def download_file(self):
        content = self.textarea.value or ''
        if not content:
            ui.notify('Nothing to download!', type='warning')
            return
        ui.download(content.encode('utf-8'), 'biblemate_notes.md')
        ui.notify('Downloaded!', type='positive')
    
    async def handle_upload(self, e):
        try:
            content = await e.file.read()
            self.textarea.value = content.decode('utf-8')
            if not self.is_editing: self.update_preview()
            ui.notify('Loaded!', type='positive')
            self.upload.reset()
        except Exception as ex:
            ui.notify(f'Error: {str(ex)}', type='negative')
    
    def clear_text(self):
        self.textarea.value = ''
        self.html_view.content = ''
        ui.notify('Cleared!', type='info')

def notepad(gui=None, q='', **_):

    def bcv(event):
        nonlocal gui
        b, c, v, *_ = event.args
        gui.change_area_1_bible_chapter(None, b, c, v)

    ui.on('bcv', bcv)

    notepad = Notepad(q)
    notepad.setup_ui()



from agentmake import agentmake, DEFAULT_AI_BACKEND
from agentmake.utils.text_wrapper import get_stream_event_text
from agentmake.utils.read_assistant_response import is_openai_style
from nicegui import ui, app, run
import asyncio, datetime, re
from biblemateweb.pages.ai.stream import stream_response
from biblemateweb import config, get_translation

def ai_chat(gui=None, q="", **_):

    SEND_BUTTON = None
    REQUEST_INPUT = None
    MESSAGES = None
    MESSAGE_CONTAINER = None
    SCROLL_AREA = None
    CANCEL_EVENT = None
    CANCEL_NOTIFICATION = None
    DELETE_DIALOG = None

    def reset_ui():
        nonlocal SEND_BUTTON, REQUEST_INPUT, CANCEL_EVENT, CANCEL_NOTIFICATION
        if not CANCEL_EVENT.is_set():
            CANCEL_EVENT.set()
        if CANCEL_EVENT is not None:
            CANCEL_EVENT = None
        SEND_BUTTON.set_text(get_translation("Send"))
        SEND_BUTTON.props('color=primary')
        REQUEST_INPUT.enable()
        REQUEST_INPUT.set_value("")
        REQUEST_INPUT.run_method('focus')
        if CANCEL_NOTIFICATION is not None:
            CANCEL_NOTIFICATION.dismiss()
            CANCEL_NOTIFICATION = None

    async def stop_confirmed():
        nonlocal DELETE_DIALOG, CANCEL_NOTIFICATION, CANCEL_EVENT
        DELETE_DIALOG.close()
        CANCEL_NOTIFICATION = ui.notification(get_translation("Stopping..."), timeout=None, spinner=True)
        if CANCEL_EVENT is not None and not CANCEL_EVENT.is_set():
            CANCEL_EVENT.set()
        await asyncio.sleep(0)

    async def handle_send_click():

        """Handles the logic when the Send button is pressed."""
        nonlocal REQUEST_INPUT, SCROLL_AREA, MESSAGE_CONTAINER, SEND_BUTTON, MESSAGES, CANCEL_EVENT
        if CANCEL_EVENT is None or CANCEL_EVENT.is_set():
            CANCEL_EVENT = asyncio.Event() # do not use threading.Event() in this case
        else:
            DELETE_DIALOG.open()
            return None

        if not MESSAGES:
            MESSAGES = [{"role": "system", "content": "You are BibleMate AI, an autonomous agent designed to assist users with their Bible study."}]
        prompt_markdown = None
        output_markdown = None
        if user_request := REQUEST_INPUT.value:
            REQUEST_INPUT.disable()
            SEND_BUTTON.set_text(get_translation("Stop"))
            SEND_BUTTON.props('color=negative')

            with MESSAGE_CONTAINER:
                ui.chat_message(user_request,
                    #name='Eliran Wong',
                    stamp=datetime.datetime.now().strftime("%H:%M"),
                    avatar='https://avatars.githubusercontent.com/u/25262722?s=96&v=4',
                )

                # when prompt-engineering is enabled
                if app.storage.user["prompt_engineering"]:
                    # holder for prompt text
                    with ui.expansion(get_translation("Prompt Engineering"), icon='tune', value=True) \
                            .classes('w-full border rounded-lg shadow-sm') \
                            .props('header-class="font-bold text-lg text-secondary"') as prompt_expansion:
                        prompt_markdown = ui.markdown().style('font-size: 1.1rem')
                    user_request = await stream_response(MESSAGES, user_request, prompt_markdown, CANCEL_EVENT, system="improve_prompt_2", scroll_area=SCROLL_AREA)
                    if not user_request or user_request.strip() == "[NO_CONTENT]":
                        reset_ui()
                        return None
                    else:
                        # refine response
                        if "```" in user_request:
                            user_request = re.sub(r"^.*?(```improved_prompt|```)(.+?)```.*?$", r"\2", user_request, flags=re.DOTALL).strip()
                            prompt_markdown.content = user_request
                            await asyncio.sleep(0)
                        # close prompt expansion
                        prompt_expansion.close()
                
                # generate response
                try:
                    with ui.expansion(get_translation("Agent"), icon='support_agent', value=True) \
                            .classes('w-full border rounded-lg shadow-sm') \
                            .props('header-class="font-bold text-lg text-secondary"') as agent_expansion:
                        agent_markdown = ui.markdown().style('font-size: 1.1rem')
                    output_markdown = ui.markdown().style('font-size: 1.1rem')
                    answers = await stream_response(MESSAGES, user_request, output_markdown, CANCEL_EVENT, system="auto" if app.storage.user["use_agent"] else None, scroll_area=SCROLL_AREA, agent_expansion=agent_expansion, agent_markdown=agent_markdown)
                    # update
                    if not answers or answers.strip() == "[NO_CONTENT]":
                        reset_ui()
                        return None
                    else:
                        MESSAGES += [
                            {"role": "user", "content": user_request},
                            {"role": "assistant", "content": answers},
                        ]
                except Exception as e:
                    output_markdown.content = f"[{get_translation('Error')}: {str(e)}]"
                    await asyncio.sleep(0)
                    import traceback
                    traceback.print_exc()
                #reset
                reset_ui()

    with ui.column().classes('w-full h-screen no-wrap gap-0') as chat_container:

        def check_screen(ev):
            nonlocal gui, chat_container
            if gui.is_portrait and app.storage.user["layout"] == 2:
                chat_container.classes(remove='h-screen', add='h-[50vh]')
            else:
                chat_container.classes(remove='h-[50vh]', add='h-screen')
            chat_container.update()

        # check screen when loaded
        check_screen(None)
        # bind
        ui.on('resize', check_screen)

        # Capture the ui.scroll_area instance in the global variable
        # w-full flex-grow p-4 border border-gray-300 rounded-lg mb-2
        with ui.column().classes('w-full flex-grow overflow-hidden'):
            with ui.scroll_area().classes('w-full p-4 border rounded-lg h-full') as SCROLL_AREA:
                MESSAGE_CONTAINER = ui.column().classes('w-full items-start gap-2')

        with ui.row().classes('w-full flex-nowrap items-end mb-0'):
            REQUEST_INPUT = ui.textarea(placeholder=get_translation("Enter your message...")).props('rows=4').classes('flex-grow h-full resize-none').on('keydown.shift.enter.prevent', handle_send_click)
            with ui.column().classes('h-full justify-between gap-2'):
                ui.checkbox(get_translation("Auto-scroll"), value=True).classes('w-full').bind_value(app.storage.user, 'auto_scroll').props('dense')
                SEND_BUTTON = ui.button('Send', on_click=handle_send_click).classes('w-full')
        
        with ui.row().classes('w-full flex-nowrap items-end mb-30'):
            ui.checkbox(get_translation("Improve Prompt"), value=True).classes('w-full').bind_value(app.storage.user, 'prompt_engineering').props('dense')
            ui.checkbox(get_translation("Agent"), value=True).classes('w-full').bind_value(app.storage.user, 'use_agent').props('dense')

        ui.label('BibleMate AI | © 2025 | Eliran Wong')

        # Dialog to confirm the reset
        with ui.dialog() as DELETE_DIALOG, ui.card():
            ui.label(f'{get_translation("Stop running task?")}?')
            with ui.row().classes('justify-end w-full'):
                ui.button(get_translation("Cancel"), on_click=DELETE_DIALOG.close).props('flat text-color=secondary')
                ui.button(get_translation("Stop"), color='negative', on_click=stop_confirmed)

        if q:
            if not q.strip().endswith("# Query"):
                q = f'# Selected text\n\n{q}\n\n# Query\n\n'
            elif q.endswith("# Query"):
                q += "\n\n"
            REQUEST_INPUT.set_value(q)
        REQUEST_INPUT.run_method('focus')
        ui.run_javascript(f'''
            const el = document.getElementById({REQUEST_INPUT.id}).querySelector('textarea');
            el.setSelectionRange(el.value.length, el.value.length);
        ''')



from pathlib import Path
from agentmake import readTextFile, writeTextFile
from agentmake.plugins.uba.lib.BibleBooks import BibleBooks
from biblemateweb import config
from biblemateweb.translations.eng import translation_eng
from biblemateweb.translations.tc import translation_tc
from biblemateweb.translations.sc import translation_sc
from nicegui import app, ui
from typing import List
import os, glob, apsw, re, asyncio
import numpy as np
import json


BIBLEMATEWEB_APP_DIR = os.path.dirname(os.path.realpath(__file__))
BIBLEMATEWEB_USER_DIR = os.path.join(os.path.expanduser("~"), "biblemate")
BIBLEMATEWEB_DATA = os.path.join(os.path.expanduser("~"), "biblemate", "data")
if not os.path.isdir(BIBLEMATEWEB_USER_DIR):
    Path(BIBLEMATEWEB_USER_DIR).mkdir(parents=True, exist_ok=True)
BIBLEMATEWEB_DATA_CUSTOM = os.path.join(os.path.expanduser("~"), "biblemate", "data_custom")
if not os.path.isdir(BIBLEMATEWEB_DATA_CUSTOM):
    Path(BIBLEMATEWEB_DATA_CUSTOM).mkdir(parents=True, exist_ok=True)
for i in ("audio", "bibles"):
    if not os.path.isdir(os.path.join(BIBLEMATEWEB_DATA, i)):
        Path(os.path.join(BIBLEMATEWEB_DATA, i)).mkdir(parents=True, exist_ok=True)
CONFIG_FILE_BACKUP = os.path.join(BIBLEMATEWEB_USER_DIR, "biblemateweb.config")

# NOTE: When add a config item, update both `write_user_config` and `default_config`

def write_user_config():
    """Writes the current configuration to the user's config file."""
    configurations = f"""config.hot_reload={config.hot_reload}
config.avatar="{config.avatar}"
config.ai_backend="{config.ai_backend}"
config.embedding_model="{config.embedding_model}"
config.custom_token="{config.custom_token}"
config.google_client_id="{config.google_client_id}"
config.google_client_secret="{config.google_client_secret}"
config.auth_uri="{config.auth_uri}"
config.storage_secret="{config.storage_secret}"
config.port={config.port}
config.verses_limit={config.verses_limit}"""
    writeTextFile(CONFIG_FILE_BACKUP, configurations)

# restore config backup after upgrade
default_config = '''config.hot_reload=False
config.avatar=""
config.ai_backend=""
config.embedding_model="paraphrase-multilingual"
config.custom_token=""
config.google_client_id=""
config.google_client_secret=""
config.auth_uri=""
config.storage_secret="REPLACE_ME_WITH_A_REAL_SECRET"
config.port=33355
config.verses_limit=2000'''

def load_config():
    """Loads the user's configuration from the config file."""
    if not os.path.isfile(CONFIG_FILE_BACKUP):
        exec(default_config, globals())
        write_user_config()
    else:
        exec(readTextFile(CONFIG_FILE_BACKUP), globals())
    # check if new config items are added
    changed = False
    for config_item in default_config[7:].split("\nconfig."):
        key, _ = config_item.split("=", 1)
        if not hasattr(config, key):
            exec(f"config.{config_item}", globals())
            changed = True
    if changed:
        write_user_config()

# load user config at startup
load_config()

# frequently used functions

class VerseEventObj:
    def __init__(self, args):
        *_, b, c, v = args
        self.args = (b, c, v)

def get_translation(text: str):
    if app.storage.user["ui_language"] == "tc":
        return translation_tc.get(text, text)
    elif app.storage.user["ui_language"] == "sc":
        return translation_sc.get(text, text)
    return translation_eng.get(text, text)

async def loading(func, *args, **kwargs):
    n = ui.notification(timeout=None)
    try:
        awaitable = asyncio.to_thread(func, *args, **kwargs)
        task = asyncio.create_task(awaitable)
        while not task.done():
            n.message = get_translation('Loading...')
            n.spinner = True
            await asyncio.sleep(0.2)
        #n.message = 'Done!'
        n.spinner = False
        await asyncio.sleep(1)
        #n.dismiss()
        return task.result()
    except Exception as e:
        n.message = f'Error: {str(e)}'
        n.type = 'negative'    
    finally:
        # 3. Always dismiss the notification, even if errors occur
        n.dismiss()

def load_topic_vectors_from_db(db_file, sql_table):
    entries = []
    entry_vectors = []
    
    with apsw.Connection(db_file) as connection:
        cursor = connection.cursor()
        cursor.execute(f"SELECT entry, entry_vector FROM {sql_table}")
        
        # Heavy CPU work: Parsing JSON
        for entry, vector_json in cursor.fetchall():
            if entry and vector_json:
                entries.append(entry)
                entry_vectors.append(np.array(json.loads(vector_json)))

    if not entries:
        return [], None

    # Heavy CPU work: Stacking arrays
    document_matrix = np.vstack(entry_vectors)
    return entries, document_matrix

def load_vectors_from_db(db_file, sql_table):
    entries = []
    entry_vectors = []
    
    # Open a NEW connection inside the worker (essential for thread safety)
    with apsw.Connection(db_file) as connection:
        cursor = connection.cursor()
        cursor.execute(f"SELECT path, entry, entry_vector FROM {sql_table}")
        
        # Heavy CPU work: Fetching and parsing JSON
        for path, entry, vector_json in cursor.fetchall():
            if path and entry and vector_json:
                entries.append(f"[{path}] {entry}")
                entry_vectors.append(np.array(json.loads(vector_json)))

    if not entries:
        return [], None

    # Heavy CPU work: Stacking arrays
    document_matrix = np.vstack(entry_vectors)
    return entries, document_matrix

# bibles resources
def getBibleInfo(db):
    abb = os.path.basename(db)[:-6]
    try:
        with apsw.Connection(db) as connn:
            query = "SELECT Title FROM Details limit 1"
            cursor = connn.cursor()
            cursor.execute(query)
            info = cursor.fetchone()
    except:
        try:
            with apsw.Connection(db) as connn:
                query = "SELECT Scripture FROM Verses WHERE Book=? AND Chapter=? AND Verse=? limit 1"
                cursor = connn.cursor()
                cursor.execute(query, (0, 0, 0))
                info = cursor.fetchone()
        except:
            return abb
    return info[0] if info else abb

bibles_dir = os.path.join(BIBLEMATEWEB_DATA, "bibles")
if os.path.isdir(bibles_dir):
    config.bibles = dict(sorted({os.path.basename(i)[:-6]: (getBibleInfo(i), i) for i in glob.glob(os.path.join(bibles_dir, "*.bible")) if not re.search("(MOB|MIB|MAB|MTB|MPB).bible$", i)}.items()))
else:
    Path(bibles_dir).mkdir(parents=True, exist_ok=True)
    config.bibles = {}
bibles_dir_custom = os.path.join(BIBLEMATEWEB_DATA_CUSTOM, "bibles")
if os.path.isdir(bibles_dir_custom):
    config.bibles_custom = dict(sorted({os.path.basename(i)[:-6]: (getBibleInfo(i), i) for i in glob.glob(os.path.join(bibles_dir_custom, "*.bible")) if not re.search("(MOB|MIB|MAB|MTB|MPB).bible$", i)}.items()))
else:
    Path(bibles_dir_custom).mkdir(parents=True, exist_ok=True)
    config.bibles_custom = {}

def getBibleVersionList(custom: bool = False) -> List[str]:
    """Returns a list of available Bible versions"""
    bibleVersionList = ["ORB", "OIB", "OPB", "ODB", "OLB"]+list(config.bibles.keys())
    if custom: # app.storage.client["custom"]
        bibleVersionList += list(config.bibles_custom.keys())
        bibleVersionList = list(set(bibleVersionList))
    return sorted(bibleVersionList)

def getBibleVersionName(abb):
    if abb in config.bibles_custom:
        return config.bibles_custom[abb][0]
    elif abb in config.bibles:
        return config.bibles[abb][0]
    return abb

def get_default_bible(language):
    if language == 'tc':
        module = "CUV"
    elif language == 'sc':
        module = "CUVs"
    else:
        module = "NET"
    return module

def resolve_verses_additional_options(q: str = None, default_bible: str = "NET", custom: bool = False):
    BIBLE_BOOKS = [BibleBooks.abbrev["eng"][str(i)][0] for i in range(1,67)]
    OT_BOOKS = BIBLE_BOOKS[:39]
    NT_BOOKS = BIBLE_BOOKS[39:]
    client_bibles = getBibleVersionList(custom)
    selected_bibles = default_bible
    selected_books = ['All', 'OT', 'NT'] + BIBLE_BOOKS
    if q and ":::" in q:
        additional_options, q = q.split(":::", 1)
        valid_books = []
        valid_bibles = []
        for i in additional_options.split(","):
            if i.strip() in selected_books+["None"]:
                valid_books.append(i.strip())
            elif i.strip() in client_bibles:
                valid_bibles.append(i.strip())
        # refine valid books and bibles
        if "None" in valid_books:
            valid_books = ["None"]
        elif "All" in valid_books or ("OT" in valid_books and "NT" in valid_books):
            valid_books = selected_books
        else:
            if "OT" in valid_books:
                valid_books += OT_BOOKS
                valid_books = list(set(valid_books))
            if "NT" in valid_books:
                valid_books += NT_BOOKS
                valid_books = list(set(valid_books))
        if valid_books:
            selected_books = valid_books
        if valid_bibles:
            selected_bibles = valid_bibles
    return client_bibles, selected_bibles, selected_books, q

def update_verses_sql_query(selected_values):
    """Generates the SQLite query based on selection."""
    BIBLE_BOOKS = [BibleBooks.abbrev["eng"][str(i)][0] for i in range(1,67)]
    BOOK_MAP = {book: i + 1 for i, book in enumerate(BIBLE_BOOKS)}
    OT_BOOKS = BIBLE_BOOKS[:39]
    NT_BOOKS = BIBLE_BOOKS[39:]
    if "OT" in selected_values:
        selected_values.remove("OT")
        selected_values += OT_BOOKS
    if "NT" in selected_values:
        selected_values.remove("NT")
        selected_values += NT_BOOKS
    selected_values = list(set(selected_values))
    
    # Filter to keep ONLY the actual book strings (ignore All/None/OT/NT)
    real_books = [b for b in selected_values if b in BIBLE_BOOKS]
    book_ids = [str(BOOK_MAP[b]) for b in real_books]
    
    base_query = "PRAGMA case_sensitive_like = false; SELECT * FROM Verses"
    where_clauses = []

    # Handle Book Logic
    if 'All' in selected_values:
        pass 
    elif not real_books:
        where_clauses.append("1=0")
    elif len(real_books) == 1:
        where_clauses.append(f"Book={book_ids[0]}")
    else:
        # Optimization: check if it's exactly OT or NT for cleaner SQL?
        # (Optional, but strictly sticking to IDs is safer for the engine)
        where_clauses.append(f"Book IN ({', '.join(book_ids)})")

    where_clauses.append(f"(Scripture REGEXP ?) ORDER BY Book, Chapter, Verse")

    # Assemble
    full_query = base_query
    if where_clauses:
        full_query += " WHERE " + " AND ".join(where_clauses)
    
    return full_query

# commentaries resources

config.commentaries_names = {
    "Barnes": "Notes on the Old and New Testaments (Barnes) [26 vol.]",
    "Benson": "Commentary on the Old and New Testaments (Benson) [5 vol.]",
    "BI": "Biblical Illustrator (Exell) [58 vol.]",
    "Brooks": "Complete Summary of the Bible (Brooks) [2 vol.]",
    "Calvin": "John Calvin's Commentaries (Calvin) [22 vol.]",
    "Clarke": "Commentary on the Bible (Clarke) [6 vol.]",
    "CBSC": "Cambridge Bible for Schools and Colleges (Cambridge) [57 vol.]",
    "CECNT": "Critical And Exegetical Commentary on the NT (Meyer) [20 vol.]",
    "CGrk": "Cambridge Greek Testament for Schools and Colleges (Cambridge) [21 vol.]",
    "CHP": "Church Pulpit Commentary (Nisbet) [12 vol.]",
    "CPBST": "College Press Bible Study Textbook Series (College) [59 vol.]",
    "EBC": "Expositor's Bible Commentary (Nicoll) [49 vol.]",
    "ECER": "Commentary for English Readers (Ellicott) [8 vol.]",
    "EGNT": "Expositor's Greek New Testament (Nicoll) [5 vol.]",
    "GCT": "Greek Testament Commentary (Alford) [4 vol.]",
    "Gill": "Exposition of the Entire Bible (Gill) [9 vol.]",
    "Henry": "Exposition of the Old and New Testaments (Henry) [6 vol.]",
    "HH": "Horæ Homileticæ (Simeon) [21 vol.]",
    "ICCNT": "International Critical Commentary, NT (1896-1929) [16 vol.]",
    "JFB": "Jamieson, Fausset, and Brown Commentary (JFB) [6 vol.]",
    "KD": "Commentary on the Old Testament (Keil & Delitzsch) [10 vol.]",
    "Lange": "Commentary on the Holy Scriptures: Critical, Doctrinal, and Homiletical (Lange) [25 vol.]",
    "MacL": "Expositions of Holy Scripture (MacLaren) [32 vol.]",
    "PHC": "Preacher's Complete Homiletical Commentary (Exell) [37 vol.]",
    "Pulpit": "Pulpit Commentary (Spence) [23 vol.]",
    "Rob": "Word Pictures in the New Testament (Robertson) [6 vol.]",
    "Spur": "Spurgeon's Expositions on the Bible (Spurgeon) [3 vol.]",
    "Vincent": "Word Studies in the New Testament (Vincent) [4 vol.]",
    "Wesley": "John Wesley's Notes on the Whole Bible (Wesley) [3 vol.]",
    "Whedon": "Commentary on the Old and New Testaments (Whedon) [14 vol.]",
}

def getCommentaryInfo(db):
    abb = os.path.basename(db)[1:-11]
    if abb in config.commentaries_names:
        return config.commentaries_names[abb]
    try:
        with apsw.Connection(db) as connn:
            query = "SELECT Title FROM Details limit 1"
            cursor = connn.cursor()
            cursor.execute(query)
            info = cursor.fetchone()
    except:
        return abb
    return info[0] if info else abb

commentaries_dir = os.path.join(BIBLEMATEWEB_DATA, "commentaries")
if os.path.isdir(commentaries_dir):
    config.commentaries = dict(sorted({os.path.basename(i)[1:-11]: (getCommentaryInfo(i), i) for i in glob.glob(os.path.join(commentaries_dir, "*.commentary"))}.items()))
else:
    Path(commentaries_dir).mkdir(parents=True, exist_ok=True)
    config.commentaries = {}
commentaries_dir_custom = os.path.join(BIBLEMATEWEB_DATA_CUSTOM, "commentaries")
if os.path.isdir(commentaries_dir_custom):
    config.commentaries_custom = dict(sorted({os.path.basename(i)[1:-11]: (getCommentaryInfo(i), i) for i in glob.glob(os.path.join(commentaries_dir_custom, "*.commentary"))}.items()))
else:
    Path(commentaries_dir_custom).mkdir(parents=True, exist_ok=True)
    config.commentaries_custom = {}

def getCommentaryVersionList(custom: bool = False) -> List[str]:
    """Returns a list of available Commentary versions"""
    commentaryVersionList = list(config.commentaries.keys())
    if custom: # app.storage.client["custom"]
        commentaryVersionList += list(config.commentaries_custom.keys())
        commentaryVersionList = list(set(commentaryVersionList))
    return sorted(commentaryVersionList)

def getCommentaryVersionName(abb):
    if abb in config.commentaries_custom:
        return config.commentaries_custom[abb][0]
    elif abb in config.commentaries:
        return config.commentaries[abb][0]
    return abb

# lexicons resources
lexicons_dir = os.path.join(BIBLEMATEWEB_DATA, "lexicons")
if os.path.isdir(lexicons_dir):
    config.lexicons = dict(sorted({os.path.basename(i)[:-8]: i for i in glob.glob(os.path.join(lexicons_dir, "*.lexicon"))}.items()))
else:
    Path(lexicons_dir).mkdir(parents=True, exist_ok=True)
    config.lexicons = {}
lexicons_dir_custom = os.path.join(BIBLEMATEWEB_DATA_CUSTOM, "lexicons")
if os.path.isdir(lexicons_dir_custom):
    config.lexicons_custom = dict(sorted({os.path.basename(i)[:-8]: i for i in glob.glob(os.path.join(lexicons_dir_custom, "*.lexicon"))}.items()))
else:
    Path(lexicons_dir_custom).mkdir(parents=True, exist_ok=True)
    config.lexicons_custom = {}

def getLexiconList(custom: bool = False) -> List[str]:
    """Returns a list of available Lexicons"""
    client_lexicons = list(config.lexicons.keys())
    if custom: # app.storage.client["custom"]
        client_lexicons += list(config.lexicons_custom.keys())
    return sorted(list(set(client_lexicons)))

# fonts
app.add_static_files('/fonts', os.path.join(BIBLEMATEWEB_APP_DIR, "fonts"))

# images
app.add_static_files('/timelines', os.path.join(BIBLEMATEWEB_DATA, "books", "Timelines"))

# audio resources
app.add_media_files('/bhs5_audio', os.path.join(BIBLEMATEWEB_DATA, "audio", "bibles", "BHS5", "default"))
app.add_media_files('/ognt_audio', os.path.join(BIBLEMATEWEB_DATA, "audio", "bibles", "OGNT", "default"))
audio_dir = os.path.join(BIBLEMATEWEB_DATA, "audio", "bibles")
if os.path.isdir(audio_dir):
    config.audio = {i: os.path.join(audio_dir, i, "default") for i in os.listdir(audio_dir) if os.path.isdir(os.path.join(audio_dir, i)) and not i in ("BHS5", "OGNT")}
else:
    Path(audio_dir).mkdir(parents=True, exist_ok=True)
    config.audio = {}
audio_dir_custom = os.path.join(BIBLEMATEWEB_DATA_CUSTOM, "audio", "bibles")
if os.path.isdir(audio_dir_custom):
    config.audio_custom = {i: os.path.join(audio_dir_custom, i, "default") for i in os.listdir(audio_dir_custom) if os.path.isdir(os.path.join(audio_dir_custom, i)) and not i in ("BHS5", "OGNT")}
else:
    Path(audio_dir_custom).mkdir(parents=True, exist_ok=True)
    config.audio_custom = {}

def getAudioVersionList(custom: bool = False) -> List[str]:
    """Returns a list of available Audio versions"""
    version_options = list(config.audio.keys())
    if custom:
        version_options += list(config.audio_custom.keys())
        version_options = list(set(version_options))
    return version_options

config.topics = {
    "HIT": "Hitchcock's New and Complete Analysis of the Bible",
    "NAV": "Nave's Topical Bible",
    "TCR": "Thompson Chain References",
    "TOR": "Torrey's New Topical Textbook",
    "TOP": "Topical Bible",
}
config.dictionaries = {
    "AMT": "American Tract Society Dictionary",
    "BBD": "Bridgeway Bible Dictionary",
    "BMC": "Freeman's Handbook of Bible Manners and Customs",
    "BUC": "Buck's A Theological Dictionary",
    "CBA": "Companion Bible Appendices",
    "DRE": "Dictionary Of Religion And Ethics",
    "EAS": "Easton's Illustrated Bible Dictionary",
    "FAU": "Fausset's Bible Dictionary",
    "FOS": "Bullinger's Figures of Speech",
    "HBN": "Hitchcock's Bible Names Dictionary",
    "MOR": "Morrish's Concise Bible Dictionary",
    "PMD": "Poor Man's Dictionary",
    "SBD": "Smith's Bible Dictionary",
    "USS": "Annals of the World",
    "VNT": "Vine's Expository Dictionary of New Testament Words",
}
config.encyclopedias = {
    "DAC": "Hasting's Dictionary of the Apostolic Church",
    "DCG": "Hasting's Dictionary Of Christ And The Gospels",
    "HAS": "Hasting's Dictionary of the Bible",
    "ISB": "International Standard Bible Encyclopedia",
    "KIT": "Kitto's Cyclopedia of Biblical Literature",
    "MSC": "McClintock & Strong's Cyclopedia of Biblical Literature",
}

# User Default Settings

USER_DEFAULT_SETTINGS = {
    'font_size': 100,
    'primary_color': '#827e67', # #827e67
    'secondary_color': '#c7c279', # #c7c279
    'negative_color': '#ff384f',
    'avatar': '',
    'custom_token': '',
    'primary_bible': 'NET',
    'secondary_bible': 'KJV',
    'favorite_commentary': 'AIC',
    'favorite_encyclopedia': 'ISB',
    'favorite_lexicon': 'Morphology',
    'hebrew_lexicon': 'TBESH',
    'greek_lexicon': 'TBESG',
    'ai_backend': 'googleai',
    'api_endpoint': '',
    'api_key': '',
    'ui_language': 'eng',
    'dark_mode': True,
    'notes': True,
    'left_drawer_open': False,
    'search_case_sensitivity': False,
    'search_mode': 1,
    'top_similar_entries': 5,
    'top_similar_verses': 20,
    'default_number_of_tabs1': 3,
    'default_number_of_tabs2': 3,
    'layout_swap_button': True,
    'bible_select_button': True,
    'loop_audio': True,
    'loop_podcast': True,
    'prompt_engineering': True,
    'use_agent': True,
    'auto_scroll': True,
}


translation_tc = {
    "Add Bible Tab": "新聖經標籤頁",
    "Add Parallel Tab": "新平行標籤頁",
    "Add Tool Tab": "新工具標籤頁",
    "Agent Mode": "代理模式",
    "Agent": "代理",
    "AI Chat": "對話模式",
    "AI Commentary": "AI 註釋",
    "Ask BibleMate": "問 BibleMate",
    "Bible Audio": "聖經語音",
    "Bible Chronology": "聖經時序",
    "Bible Commentaries": "聖經註釋",
    "Bible Only": "只顯示聖經",
    "Bible Parallels": "平行經文",
    "Bible Podcast": "聖經廣播",
    "Bible Promises": "聖經應許",
    "Bible": "聖經",
    "Bibles": "聖經",
    "Parallel Bibles": "平行聖經",
    "Timelines": "聖經時間軸",
    "Bible Timelines": "聖經時間軸",
    "Characters": "聖經人物",
    "Close Bible Tab": "關閉聖經標籤頁",
    "Close Others": "關閉其他",
    "Close Parallel Tab": "關閉平行標籤頁",
    "Close Tool Tab": "關閉工具標籤頁",
    "Commentaries": "聖經註釋",
    "Comparison": "版本對比",
    "Copy": "複製",
    "Cross-references": "經文串珠",
    "Dictionaries": "聖經字典",
    "Discourse Analysis": "語法分析",
    "Encyclopedias": "百科全書",
    "Go": "開啓",
    "Indexes": "資源索引",
    "Chapter Indexes": "全章索引",
    "Verse Indexes": "單節索引",
    "Lexicons": "原文詞典",
    "Locations": "聖經地點",
    "Maps": "聖經地圖",
    "Morphological Data": "形態學數據",
    "Morphology": "形態學分析",
    "Names": "聖經名字",
    "Parallels": "平行經文",
    "Partner Mode": "搭檔模式",
    "Preferences": "偏好設定",
    "Promises": "聖經應許",
    "Relationships": "人物關係",
    "Screen": "全螢幕",
    "Fullscreen": "全螢幕",
    "Swap": "版面轉換",
    "Sync": "同步",
    "Tool Only": "只顯示工具",
    "Topics": "聖經主題",
    "Treasury": "經文寶庫",
    "Treasury of Scripture Knowledge": "經文寶庫",
    "Bible & Tool": "聖經與工具",
    "Light Mode": "淺色模式",
    "Dark Mode": "深色模式",
    "Verses": "搜索經文",
    "Tools": "工具",
    "Search": "搜索",
    "Quick search": "快速搜索",
    "AI": "人工智能",
    "Settings": "設定",
    "Swap Layout": "版面轉換",
    "Select Bible Verse": "聖經選單",
    "Toggle Display of Swap Layout Button": "[開啓/關閉] 版面轉換按鈕",
    "Toggle Display of Bible Selection Button": "[開啓/關閉] 聖經選單按鈕",
    "Toggle Bible Synchronization": "[開啓/關閉] 聖經同步功能",
    "Toggle Fullscreen": "[開啓/關閉] 全螢幕",
    "Prev Chapter": "上一章",
    "Next Chapter": "下一章",
    "OT": "舊約",
    "NT": "新約",
    "Select Book": "選擇書卷",
    "Select Chapter": "選擇章數",
    "Select Verse": "選擇節數",
    "Back": "返回",
    "Share": "分享",
    "Note": "筆記",
    "Notes": "筆記",
    "Bible Notes": "聖經筆記",
    "Notepad": "筆記本",
    "Add Note": "新增筆記",
    "Edit Note": "編輯筆記",
    "Save": "儲存",
    "Download": "下載",
    "Import": "匯入",
    "Delete": "刪除",
    "Logout": "登出",
    "Read": "閱讀",
    "Edit": "編輯",
    "Clear": "清除",
    "Download Index": "下載索引",
    "Rebuild Index": "重建索引",
    "Download All": "下載所有筆記",
    "Book Note": "書卷筆記",
    "Chapter Note": "全章筆記",
    "Verse Note": "經文筆記",
    "Book Tools": "書卷工具",
    "Chapter Tools": "全章工具",
    "Verse Tools": "單節工具",
    "Chapter Summary": "全章摘要",
    "Verse Commentary": "經文註釋",
    "Book Analysis": "書卷分析",
    "Overview": "概述",
    "Structural Outline": "結構大綱",
    "Logical Flow": "書卷脈絡",
    "Historical Setting": "歷史背景",
    "Themes": "主題",
    "Keywords": "關鍵詞",
    "Theology": "神學",
    "Canonical Placement": "正典位置",
    "Practical Living": "生活應用",
    "Summary": "摘要",
    "Chapter Map": "全章地圖",
    "Loop": "循環",
    "Edit Mode": "編輯模式",
    "Read Mode": "閱讀模式",
    "Literal": "字面",
    "Regex": "模式",
    "Semantic": "語義",
    "Case-sensitive": "大小寫",
    "Type to filter": "輸入以篩選",
    "results": "結果",
    "result": "結果",
    "Search for words or refs": "搜索經文或參照（如：申 6:4; 約 3:16-18）",
    "Search Bible dictionaries": "搜索聖經字典",
    "Search for a bible topic": "搜索聖經主題",
    "Search for bible promises": "搜索聖經應許",
    "Search for parallel passages": "搜索平行經文",
    "Search for a name or meaning": "搜索名字或含義",
    "Search for location(s) or verse reference(s)": "搜索地點或經文參照",
    "Search for a bible location": "搜索聖經地點",
    "Search for a bible character": "搜索聖經人物",
    "Show Verses": "顯示所有經文",
    "Distance": "距離",
    "From": "從",
    "To": "到",
    "km": "公里",
    "miles": "英里",
    "Loading...": "載入中...",
    "Connecting...": "連線中...",
    "Canceling...": "取消中...",
    "Running...": "執行中...",
    "Stopping...": "停止中...",
    "Stop running task?": "停止執行中的工作？",
    "Enter bible verse reference(s) here...": "在此輸入經文參照...",
    "Start typing your notes here...": "在此開始輸入您的筆記...",
    "Reset All Preferences": "重置所有偏好設定",
    "Language": "語言",
    "Delete": "刪除",
    "Cancel": "取消",
    "Cancelled!": "已取消！",
    "Done!": "完成！",
    "Stop": "停止",
    "Start": "開始",
    "Send": "發送",
    "Auto-scroll": "自動滾動",
    "Improve Prompt": "優化指示",
    "Semantic Searches": "語義搜索",
    "Frequently Used Resources": "常用資源",
    "User & Custom Data": "用戶與自定數據",
    "Tabs Management": "標籤頁管理",
    "Appearance": "外觀",
    "Enter your message...": "輸入您的訊息...",
    "Creating agent...": "創建代理中...",
    "Created!": "已創建！",
    "Error": "錯誤",
}


from nicegui import ui, app, run
import os, re, apsw, markdown2, asyncio
from biblemateweb import BIBLEMATEWEB_DATA, get_translation
from biblemateweb.fx.bible import BibleSelector

def fetch_summary(b, c, lang="eng"):
    if lang == "tc":
        db = os.path.join(BIBLEMATEWEB_DATA, "data", "chapter_summary_tc.data")
    elif lang == "sc":
        db = os.path.join(BIBLEMATEWEB_DATA, "data", "chapter_summary_sc.data")
    else:
        db = os.path.join(BIBLEMATEWEB_DATA, "data", "chapter_summary.data")
    fetch = None
    with apsw.Connection(db) as connn:
        query = "SELECT Content FROM Summary WHERE Book=? AND Chapter=? LIMIT 1"
        cursor = connn.cursor()
        cursor.execute(query, (b, c))
        fetch = cursor.fetchone()
    return fetch[0] if fetch else "No summary available."

def chapter_summary(gui=None, b=1, c=1, v=1, area=2, **_):

    # handle bcv events
    def bcv(event):
        nonlocal gui
        b, c, v, *_ = event.args
        gui.change_area_1_bible_chapter(None, b, c, v)
    ui.on('bcv', bcv)

    # Bible Selection menu
    bible_selector = BibleSelector(version_options=["KJV"])
    def additional_items():
        nonlocal gui, bible_selector, area
        def change_summary_chapter(selection):
            if area == 1:
                _, app.storage.user['bible_book_number'], app.storage.user['bible_chapter_number'], app.storage.user['bible_verse_number'] = selection
                gui.load_area_1_content(title="Summary")
            else:
                _, app.storage.user['tool_book_number'], app.storage.user['tool_chapter_number'], app.storage.user['tool_verse_number'] = selection
                gui.load_area_2_content(title="Summary", sync=False)
        ui.button(get_translation('Go'), on_click=lambda: change_summary_chapter(bible_selector.get_selection()))
    bible_selector.create_ui("KJV", b, c, v, additional_items=additional_items, show_versions=False, show_verses=False)

    # Summary display
    async def load_summary(b, c):
        n = ui.notification("Loading ...", timeout=None, spinner=True)
        await asyncio.sleep(0)
        # fetch content
        content = await run.io_bound(fetch_summary, b, c, app.storage.user['ui_language'])
        # clean up content
        if app.storage.user['ui_language'] == "tc":
            pattern = r'^好的，我將為[你您]\s*.*?#'
            content = re.sub(pattern, '#', content, flags=re.DOTALL)
            pattern = r'\n---\s*\n如果[你您]願意，\s*.*$'
            content = re.sub(pattern, '', content, flags=re.DOTALL)
        elif app.storage.user['ui_language'] == "sc":
            pattern = r'^好的，我将为[你您]\s*.*?#'
            content = re.sub(pattern, '#', content, flags=re.DOTALL)
            pattern = r'\n---\s*\n如果[你您]愿意，\s*.*$'
            content = re.sub(pattern, '', content, flags=re.DOTALL)
        else:
            pattern = r'^Alright\s*.*?#'
            content = re.sub(pattern, '#', content, flags=re.DOTALL)
            pattern = r'\n---\s*\nIf you\’d like\,\s*.*$'
            content = re.sub(pattern, '', content, flags=re.DOTALL)
        # convert md to html
        content = markdown2.markdown(content, extras=["tables","fenced-code-blocks","toc","codelite"])
        content = content.replace("<h1>", "<h2>").replace("</h1>", "</h2>")
        # convert links
        content = re.sub(r'''(onclick|ondblclick)="(bcv)\((.*?)\)"''', r'''\1="emitEvent('\2', [\3]); return false;"''', content)
        # display content
        ui.html(f'<div class="content-text">{content}</div>', sanitize=False)
        # dismiss notification
        n.dismiss()

    
    ui.timer(0, lambda: load_summary(b, c), once=True)   


translation_sc = {
    "Add Bible Tab": "新圣经标签页",
    "Add Parallel Tab": "新平行标签页",
    "Add Tool Tab": "新工具标签页",
    "Agent Mode": "代理模式",
    "Agent": "代理",
    "AI Chat": "对话模式",
    "AI Commentary": "AI 注释",
    "Ask BibleMate": "问 BibleMate",
    "Bible Audio": "圣经语音",
    "Bible Chronology": "圣经时序",
    "Bible Commentaries": "圣经注释",
    "Bible Only": "只显示圣经",
    "Bible Parallels": "平行经文",
    "Bible Podcast": "圣经广播",
    "Bible Promises": "圣经应许",
    "Bible": "圣经",
    "Bibles": "圣经",
    "Parallel Bibles": "平行圣经",
    "Timelines": "圣经时间轴",
    "Bible Timelines": "圣经时间轴",
    "Characters": "圣经人物",
    "Close Bible Tab": "关闭圣经标签页",
    "Close Others": "关闭其他",
    "Close Parallel Tab": "关闭平行标签页",
    "Close Tool Tab": "关闭工具标签页",
    "Commentaries": "圣经注释",
    "Comparison": "版本对比",
    "Copy": "复制",
    "Cross-references": "经文串珠",
    "Dictionaries": "圣经字典",
    "Discourse Analysis": "语法分析",
    "Encyclopedias": "百科全书",
    "Go": "开启",
    "Indexes": "资源索引",
    "Chapter Indexes": "全章索引",
    "Verse Indexes": "单节索引",
    "Lexicons": "原文词典",
    "Locations": "圣经地点",
    "Maps": "圣经地图",
    "Morphological Data": "形态学数据",
    "Morphology": "形态学分析",
    "Names": "圣经名字",
    "Parallels": "平行经文",
    "Partner Mode": "搭档模式",
    "Preferences": "偏好设定",
    "Promises": "圣经应许",
    "Relationships": "人物关系",
    "Screen": "全萤幕",
    "Fullscreen": "全萤幕",
    "Swap": "版面转换",
    "Sync": "同步",
    "Tool Only": "只显示工具",
    "Topics": "圣经主题",
    "Treasury": "经文宝库",
    "Treasury of Scripture Knowledge": "经文宝库",
    "Bible & Tool": "圣经与工具",
    "Light Mode": "浅色模式",
    "Dark Mode": "深色模式",
    "Verses": "搜索经文",
    "Tools": "工具",
    "Search": "搜索",
    "Quick search": "快速搜索",
    "AI": "人工智能",
    "Settings": "设定",
    "Swap Layout": "版面转换",
    "Select Bible Verse": "圣经选单",
    "Toggle Display of Swap Layout Button": "[开启/关闭] 版面转换按钮",
    "Toggle Display of Bible Selection Button": "[开启/关闭] 圣经选单按钮",
    "Toggle Bible Synchronization": "[开启/关闭] 圣经同步功能",
    "Toggle Fullscreen": "[开启/关闭] 全萤幕",
    "Prev Chapter": "上一章",
    "Next Chapter": "下一章",
    "OT": "旧约",
    "NT": "新约",
    "Select Book": "选择书卷",
    "Select Chapter": "选择章数",
    "Select Verse": "选择节数",
    "Back": "返回",
    "Share": "分享",
    "Note": "笔记",
    "Notes": "笔记",
    "Bible Notes": "圣经笔记",
    "Notepad": "笔记本",
    "Add Note": "新增笔记",
    "Edit Note": "编辑笔记",
    "Save": "保存",
    "Download": "下载",
    "Import": "导入",
    "Delete": "删除",
    "Logout": "登出",
    "Read": "阅读",
    "Edit": "编辑",
    "Clear": "清除",
    "Download Index": "下载索引",
    "Rebuild Index": "重建索引",
    "Download All": "下载所有笔记",
    "Book Note": "书卷笔记",
    "Chapter Note": "全章笔记",
    "Verse Note": "经文笔记",
    "Book Tools": "书卷工具",
    "Chapter Tools": "全章工具",
    "Verse Tools": "单节工具",
    "Chapter Summary": "全章摘要",
    "Verse Commentary": "经文注释",
    "Book Analysis": "书卷分析",
    "Overview": "概述",
    "Structural Outline": "结构大纲",
    "Logical Flow": "书卷脉络",
    "Historical Setting": "历史背景",
    "Themes": "主题",
    "Keywords": "关键字",
    "Theology": "神学",
    "Canonical Placement": "正典位置",
    "Practical Living": "生活应用",
    "Summary": "摘要",
    "Chapter Map": "全章地图",
    "Loop": "循环",
    "Edit Mode": "编辑模式",
    "Read Mode": "阅读模式",
    "Literal": "字面",
    "Regex": "模式",
    "Semantic": "语义",
    "Case-sensitive": "大小写",
    "Type to filter": "输入以筛选",
    "results": "结果",
    "result": "结果",
    "Search for words or refs": "搜索经文或参照（如：申 6:4; 约 3:16-18）",
    "Search Bible dictionaries": "搜索圣经字典",
    "Search for a bible topic": "搜索圣经主题",
    "Search for bible promises": "搜索圣经应许",
    "Search for parallel passages": "搜索平行经文",
    "Search for a name or meaning": "搜索名字或含义",
    "Search for location(s) or verse reference(s)": "搜索地点或经文参照",
    "Search for a bible location": "搜索圣经地点",
    "Search for a bible character": "搜索圣经人物",
    "Show Verses": "显示所有经文",
    "Distance": "距离",
    "From": "从",
    "To": "到",
    "km": "公里",
    "miles": "英里",
    "Loading...": "载入中...",
    "Connecting...": "连接中...",
    "Canceling...": "取消中...",
    "Running...": "运行中...",
    "Stopping...": "停止中...",
    "Stop running task?": "停止运行中的工作？",
    "Enter bible verse reference(s) here...": "在此输入经文参照...",
    "Start typing your notes here...": "在此开始输入您的笔记...",
    "Reset All Preferences": "重置所有偏好设定",
    "Language": "语言",
    "Delete": "删除",
    "Cancel": "取消",
    "Cancelled!": "已取消！",
    "Done!": "完成！",
    "Stop": "停止",
    "Start": "开始",
    "Send": "发送",
    "Auto-scroll": "自动滚动",
    "Improve Prompt": "优化指示",
    "Semantic Searches": "语义搜索",
    "Frequently Used Resources": "常用资源",
    "User & Custom Data": "用户与自定数据",
    "Tabs Management": "标签页管理",
    "Appearance": "外观",
    "Enter your message...": "输入您的消息...",
    "Creating agent...": "创建代理中...",
    "Created!": "已创建！",
    "Error": "错误",
}


import re, zipfile, json, io, markdown2, re, base64
import urllib.parse
from nicegui import ui, app, run
from biblemateweb import get_translation, loading
from biblemateweb.fx.bible import BibleSelector
from agentmake.plugins.uba.lib.BibleParser import BibleVerseParser
from biblemateweb.fx.cloud_index_manager import get_drive_service, CloudIndexManager
from googleapiclient.http import MediaIoBaseUpload


class CloudNotepad:
    def __init__(self, content=""):
        self.text_content = content
        self.is_editing = True
        self.parser = BibleVerseParser(False, language=app.storage.user['ui_language'])
        
    def setup_ui(self):

        # --- Content Area ---
        # Card must be 'flex flex-col' so the child (textarea) can grow
        with ui.card().classes('w-full h-[75vh] p-0 flex flex-col'):
            
            # 1. Edit Mode: Text Area
            # We apply our custom 'full-height-textarea' class here
            self.textarea = ui.textarea(
                placeholder=get_translation("Start typing your notes here..."),
                value=self.text_content
            ).classes('w-full flex-grow full-height-textarea p-2 border-none focus:outline-none') \
             .props('flat squares resize-none') \
             .bind_visibility_from(self, 'is_editing')

            # 2. Read Mode: HTML Preview
            with ui.scroll_area().classes('w-full flex-grow p-2') \
                    .bind_visibility_from(self, 'is_editing', backward=lambda x: not x):
                self.html_view = ui.html(f'<div class="content-text">{self.text_content}</div>', sanitize=False).classes('w-full prose max-w-none')

    def toggle_mode(self):
        self.is_editing = not self.is_editing
        if not self.is_editing:
            self.update_preview()

    def update_preview(self):
        content = self.textarea.value or ''
        try:
            # Added your requested extras
            html_content = markdown2.markdown(
                content, 
                extras=["tables", "fenced-code-blocks", "toc", "cuddled-lists"]
            )
            html_content = self.parser.parseText(html_content)
            html_content = re.sub(r'''(onclick|ondblclick)="(bdbid|lex|cr|bcv|website)\((.*?)\)"''', r'''\1="emitEvent('\2', [\3]); return false;"''', html_content)
            html_content = re.sub(r"""(onclick|ondblclick)='(bdbid|lex|cr|bcv|website)\((.*?)\)'""", r"""\1='emitEvent("\2", [\3]); return false;'""", html_content)
            self.html_view.content = html_content
        except Exception as e:
            self.html_view.content = f"<p class='text-red-500'>Error: {str(e)}</p>"

    def download_file(self):
        content = self.textarea.value or ''
        if not content:
            ui.notify('Nothing to download!', type='warning')
            return
        ui.download(content.encode('utf-8'), 'biblemate_notes.md')
        ui.notify('Downloaded!', type='positive')
    
    async def handle_upload(self, e):
        try:
            content = await e.file.read()
            self.textarea.value = content.decode('utf-8')
            if not self.is_editing: self.update_preview()
            ui.notify('Loaded!', type='positive')
            self.upload.reset()
        except Exception as ex:
            ui.notify(f'Error: {str(ex)}', type='negative')
    
    def clear_text(self):
        self.textarea.value = ''
        self.html_view.content = ''
        ui.notify('Cleared!', type='info')

def notes(gui=None, bt=None, b=1, c=1, v=1, area=2, **_):
    if not bt:
        bt = gui.get_area_1_bible_text()
    elif bt in ("ORB", "OIB", "OPB", "ODB", "OLB"):
        bt = "OHGB"
    
    # Book Note
    if c == 0:
        v = 0

    # 1. Auth Check
    token = app.storage.user.get('google_token', "")
    if not token:
        with ui.card().classes('absolute-center'):
            ui.html('Sign in with Google to securely save and sync your personal Bible notes across all your devices.<br><i><b>Data Policy Note:</b> BibleMate AI does not collect or store your personal notes. Your notes are saved directly within your own Google Account.</i>', sanitize=False)
            with ui.row().classes('w-full justify-center'):
                ui.button('Login with Google', on_click=lambda: ui.navigate.to('/login'))
            with ui.expansion(get_translation("BibleMate AI Data Policy & Privacy Commitment"), icon='privacy_tip').props('header-class="text-secondary"'):
                ui.html("<b>We respect your privacy.</b> BibleMate AI is designed to protect your personal data. We do not collect, store, or share your personal Bible notes. When you log in with your Google Account, your notes are created and stored exclusively on <b>your personal Google Drive/Account</b>, ensuring that you retain full control and ownership of your private data at all times.", sanitize=False)
        return

    bible_selector = None
    service = get_drive_service(token)
    index_mgr = CloudIndexManager(service)
    notepad = CloudNotepad()

    def change_note(version=None, book=1, chapter=1, verse=1):
        nonlocal gui
        _, app.storage.user['tool_book_number'], app.storage.user['tool_chapter_number'], app.storage.user['tool_verse_number'] = version, book, chapter, verse
        gui.load_area_2_content(title='Notes', sync=False)
    bible_selector = BibleSelector(version_options=[bt], on_book_changed=change_note, on_chapter_changed=change_note, on_verse_changed=change_note, chapter_zero=True, verse_zero=True)

    def refresh_ui():
        try:
            nonlocal gui, b, c, v
            active_area1_tab = gui.get_active_area1_tab()
            if active_area1_tab in app.storage.user and app.storage.user[active_area1_tab]["b"] == b and app.storage.user[active_area1_tab]["c"] == c:
                gui.change_area_1_bible_chapter(book=b, chapter=c, verse=v)
        except:
            import traceback
            traceback.print_exc()

    def get_filename(verse_id):
        return f"{verse_id}.json"

    def get_vid(): 
        return f"{b}_{c}_{v}"

    def load_current_note():
        vid = get_vid()
        try:
            filename = get_filename(vid)
            results = service.files().list(
                q=f"name='{filename}' and 'appDataFolder' in parents and trashed=false",
                spaces='appDataFolder',
                fields='files(id)'
            ).execute()
            files = results.get('files', [])
            
            if files:
                request = service.files().get_media(fileId=files[0]['id'])
                data = json.loads(request.execute())
                return data.get("content", "")
            return ""
        except Exception as e:
            ui.notify(f"Error loading: {e}", type='negative')
            return ""

    def save_note_sync(service, verse_id, content, filename):
        """
        Performs the slow network calls to Google Drive.
        Running this on the main thread would freeze the app.
        """
        import json, io
        from googleapiclient.http import MediaIoBaseUpload
        
        # Prepare the file data
        file_data = {"verse_id": verse_id, "content": content}
        media = MediaIoBaseUpload(
            io.BytesIO(json.dumps(file_data).encode('utf-8')), 
            mimetype='application/json'
        )
        
        # 1. Search for existing file
        results = service.files().list(
            q=f"name='{filename}' and 'appDataFolder' in parents and trashed=false",
            spaces='appDataFolder',
            fields='files(id)'
        ).execute()
        files = results.get('files', [])

        # 2. Update or Create
        if files:
            service.files().update(fileId=files[0]['id'], media_body=media).execute()
            return "updated"
        else:
            meta = {'name': filename, 'parents': ['appDataFolder']}
            service.files().create(body=meta, media_body=media).execute()
            return "created"

    async def save_current_note():
        # 1. Get Data
        vid = get_vid()
        content = notepad.textarea.value
        filename = get_filename(vid)
        
        # 2. Show Loading Spinner
        # (This prevents the user from clicking save twice)
        n = ui.notification('Saving to Cloud...', timeout=None, spinner=True)
        
        try:
            # 3. Run the heavy network call in a separate thread
            # 'service' is passed explicitly to be thread-safe
            result = await run.io_bound(
                save_note_sync, 
                service, 
                vid, 
                content, 
                filename
            )
            
            # 4. Update Index & UI (Back on the main thread)
            index_mgr.add_verse(vid)
            app.storage.user['cached_index'] = index_mgr.data
            
            n.dismiss()
            ui.notify('Saved successfully!', type='positive')
            refresh_ui()
            
        except Exception as e:
            n.dismiss()
            ui.notify(f"Error saving: {e}", type='negative')

    def delete_current_note():
        vid = get_vid()
        try:
            filename = get_filename(vid)
            results = service.files().list(
                q=f"name='{filename}' and 'appDataFolder' in parents and trashed=false",
                spaces='appDataFolder',
                fields='files(id)'
            ).execute()
            files = results.get('files', [])

            if files:
                service.files().delete(fileId=files[0]['id']).execute()
                index_mgr.remove_verse(vid)
                app.storage.user['cached_index'] = index_mgr.data
                notepad.textarea.value = ""
                ui.notify('Note deleted.')
                refresh_ui()
            else:
                ui.notify('Nothing to delete.')
        except Exception as e:
            ui.notify(f"Delete error: {e}", type='negative')

    async def run_rebuild():
        # Run the rebuild
        n = ui.notification('Rebuilding search index...', timeout=None, spinner=True)
        count = await run.io_bound(index_mgr.rebuild_index)
        n.dismiss()
        if count >= 0:
            app.storage.user['cached_index'] = index_mgr.data # Update session
            refresh_ui() # Refresh stats
            ui.notify(f'Search index rebuilt! Found {count} notes.', type='positive')
        else:
            ui.notify('Rebuild failed. Check console logs.', type='negative')

    def download_index_backup():
        """Generates a downloadable JSON file of the current index."""
        if not index_mgr.data:
            ui.notify('Index is empty!', type='warning')
            return
        try:
            # 1. Convert index data to pretty JSON string
            json_str = json.dumps(index_mgr.data, indent=2)
            # 2. Encode it for the browser (Data URI format)
            encoded_json = urllib.parse.quote(json_str)
            # 3. Trigger download
            ui.download(f'data:application/json,{encoded_json}', 'bible_index_backup.json')
            ui.notify('Download started.')
        except Exception as e:
            ui.notify(f'Download error: {e}', type='negative')

    # ---------------------------------------------------------
    # Import a single json or a zip
    # ---------------------------------------------------------

    def process_import_sync(file_bytes, is_zip, drive_service):
        """
        Parses the uploaded bytes (Zip or JSON) and uploads them to Drive.
        Returns a list of successfully imported verse_ids.
        """
        imported_ids = []
        files_to_process = {} # {filename: json_content_dict}

        try:
            # A. Parse the Input (Zip or Single JSON)
            if is_zip:
                with zipfile.ZipFile(io.BytesIO(file_bytes)) as z:
                    for filename in z.namelist():
                        if filename.endswith('.json') and filename != 'bible_index.json':
                            try:
                                content = z.read(filename)
                                data = json.loads(content)
                                files_to_process[filename] = data
                            except:
                                print(f"Skipping invalid file: {filename}")
            else:
                # Single JSON file
                try:
                    data = json.loads(file_bytes)
                    # Quick check if it's a note file
                    if 'verse_id' in data:
                        filename = f"{data['verse_id']}.json"
                        files_to_process[filename] = data
                except:
                    pass

            # B. Upload Loop
            # Note: This loops through every note. For 100+ notes, this takes time.
            for filename, data in files_to_process.items():
                verse_id = data.get('verse_id')
                if not verse_id: continue

                # Prepare content
                # We strictly re-dump the data to ensure clean JSON
                media = MediaIoBaseUpload(
                    io.BytesIO(json.dumps(data).encode('utf-8')), 
                    mimetype='application/json'
                )
                
                # 1. Check if file exists in Drive
                results = drive_service.files().list(
                    q=f"name='{filename}' and 'appDataFolder' in parents and trashed=false",
                    spaces='appDataFolder',
                    fields='files(id)'
                ).execute()
                existing_files = results.get('files', [])

                if existing_files:
                    # Update existing
                    file_id = existing_files[0]['id']
                    drive_service.files().update(fileId=file_id, media_body=media).execute()
                else:
                    # Create new
                    meta = {'name': filename, 'parents': ['appDataFolder']}
                    drive_service.files().create(body=meta, media_body=media).execute()
                
                imported_ids.append(verse_id)

        except Exception as e:
            print(f"Import process failed: {e}")
        
        return imported_ids

    async def handle_import(e):
        """
        UI Callback for the upload button.
        """
        #print(dir(e))
        # 1. Check login
        if not service:
            ui.notify('Please log in to Google first!', type='warning')
            return

        n = ui.notification('Processing Import... Please wait.', timeout=None, spinner=True)
        
        try:
            # 2. Read file bytes
            file_bytes = await e.file.read() 
            filename = getattr(e.file, 'name', 'unknown.json').lower()
            #print(filename)
            is_zip = filename.endswith('.zip')

            # Run background worker
            imported_ids = await run.io_bound(
                process_import_sync, 
                file_bytes, 
                is_zip, 
                service
            )

            # 4. Update Local Index
            if imported_ids:
                count = len(imported_ids)
                for vid in imported_ids:
                    index_mgr.add_verse(vid) # Updates local memory
                
                # Sync index to Drive once at the end
                index_mgr.save_to_drive() 
                app.storage.user['cached_index'] = index_mgr.data
                
                ui.notify(f'Successfully imported {count} notes!', type='positive')
                refresh_ui() # Update your stats/counters
                change_note(book=b, chapter=c, verse=v)
            else:
                ui.notify('No valid notes found to import.', type='warning')

        except Exception as err:
            ui.notify(f'Import failed: {err}', type='negative')
        finally:
            n.dismiss()

    # ---------------------------------------------------------
    # Download a single (the current) note
    # ---------------------------------------------------------
    
    def download_current_note():
        """
        Exports the currently visible note as a JSON file.
        This format is compatible with the 'Import' feature.
        """
        vid = get_vid() # Uses your existing helper
        content = notepad.textarea.value or ''
        
        if not content.strip():
            ui.notify('Cannot download an empty note.', type='warning')
            return

        try:
            # 1. Structure the data exactly like it is stored in Drive
            note_data = {
                "verse_id": vid,
                "content": content,
                "exported_at": "true" # Optional flag
            }
            
            # 2. Convert to JSON bytes
            json_str = json.dumps(note_data, indent=2)
            json_bytes = json_str.encode('utf-8')
            
            # 3. Encode to Base64 (Safe for all browsers/characters)
            b64_str = base64.b64encode(json_bytes).decode()
            
            # 4. Trigger Download
            filename = f"{vid}.json"
            ui.download(f'data:application/json;base64,{b64_str}', filename)
            ui.notify(f'Downloaded {filename}')
            
        except Exception as e:
            ui.notify(f'Download error: {e}', type='negative')

    # ---------------------------------------------------------
    # 1. The Heavy Lifter (Download ALL)
    # ---------------------------------------------------------
    def generate_zip_sync(data, drive_service):
        """
        Synchronous function to download files and create zip.
        Note: We pass 'data' and 'drive_service' explicitly.
        """
        zip_buffer = io.BytesIO()
        
        with zipfile.ZipFile(zip_buffer, 'w', zipfile.ZIP_DEFLATED) as zip_file:
            # Add Index File
            zip_file.writestr('bible_index.json', json.dumps(data, indent=2))
            
            # Loop through notes
            for verse_id in data:
                try:
                    filename = f"{verse_id}.json" # Recreating get_filename logic here for safety
                    
                    # Search logic
                    results = drive_service.files().list(
                        q=f"name='{filename}' and 'appDataFolder' in parents and trashed=false",
                        spaces='appDataFolder',
                        fields='files(id)'
                    ).execute()
                    files = results.get('files', [])

                    if files:
                        file_id = files[0]['id']
                        request = drive_service.files().get_media(fileId=file_id)
                        file_content = request.execute()
                        zip_file.writestr(filename, file_content)
                except Exception as e:
                    print(f"Error zipping {verse_id}: {e}")

        # Finalize
        zip_buffer.seek(0)
        return base64.b64encode(zip_buffer.getvalue()).decode()

    # ---------------------------------------------------------
    # 2. The UI Function (Runs on main thread)
    # ---------------------------------------------------------
    async def download_all_notes_zip():
        if not index_mgr.data:
            ui.notify('No notes to download!', type='warning')
            return

        # 1. Show Spinner (Now it will actually render!)
        n = ui.notification('Downloading notes from Google Drive... Please wait.', timeout=None, spinner=True)
        
        try:
            # 2. Offload the heavy work to a background thread
            # This allows the UI to keep running while Python waits for the result
            b64_zip = await run.io_bound(generate_zip_sync, index_mgr.data, service)
            
            # 3. Trigger Download
            ui.download(f'data:application/zip;base64,{b64_zip}', 'biblemate_notes_backup.zip')
            
            n.dismiss()
            ui.notify('Download started!', type='positive')

        except Exception as e:
            n.dismiss()
            ui.notify(f'Zip error: {e}', type='negative')
    
    # Dialog to confirm deleting a note
    with ui.dialog() as delete_dialog, ui.card():
        ui.label('Are you sure you want to delete this note?')
        with ui.row().classes('justify-end w-full'):
            ui.button('Cancel', on_click=delete_dialog.close).props('flat text-color=secondary')
            ui.button('Delete', color='red', on_click=lambda: (delete_current_note(), delete_dialog.close()))

    # Bible Selection menu
    def additional_items():
        nonlocal gui, bible_selector, area
        with ui.button(icon='more_vert').props(f'flat round color={"white" if app.storage.user["dark_mode"] else "black"}'):
            with ui.menu():
                ui.menu_item(f'👁️ {get_translation("Read")} / {get_translation("Edit")}', on_click=notepad.toggle_mode)
                ui.separator()
                ui.menu_item(f'💾 {get_translation("Save")}', on_click=save_current_note)
                ui.menu_item(f'❌ {get_translation("Delete")}', on_click=delete_dialog.open)
                ui.menu_item(f'📥 {get_translation("Download")}', on_click=download_current_note)
                ui.menu_item(f'📥 {get_translation("Download All")}', on_click=download_all_notes_zip)
                ui.separator()
                with ui.menu_item(f'📤 {get_translation("Import")}', on_click=lambda: upload_element.run_method('pickFiles')):
                    ui.tooltip('Import a single note (*.json file) or a zip file containing multiple notes.')
                ui.separator()
                with ui.menu_item(f'🛠️ {get_translation("Rebuild Index")}', on_click=run_rebuild):
                    ui.tooltip('Rebuild search index for maintenance or repair.')
                ui.menu_item(f'📥 {get_translation("Download Index")}', on_click=download_index_backup)
                ui.separator()
                with ui.menu_item(f'🔒 {get_translation("Logout")}', on_click=lambda: (app.storage.user.clear(), ui.navigate.to('/'))):
                    ui.tooltip('Log out to keep your account secure, especially on shared devices.')
    bible_selector.create_ui(bt, b, c, v, additional_items=additional_items, show_versions=False)

    upload_element = ui.upload(
        on_upload=handle_import, 
        auto_upload=True
    ).props('accept=.json,.zip').classes('hidden')

    notepad.setup_ui()
    def load_initial_content():
        notepad.textarea.value = load_current_note()
    ui.timer(0, lambda: loading(load_initial_content), once=True)




import apsw
import re, os
from nicegui import ui, app
from biblemateweb import BIBLEMATEWEB_DATA, get_translation
from functools import partial
from biblemateweb.fx.bible import BibleSelector


def resource_indexes(gui=None, bt=None, b=1, c=1, v=1, area=2, **_):

    def exlbl(event):
        nonlocal gui
        app.storage.user['tool_query'], *_ = event.args
        gui.select_empty_area2_tab()
        gui.load_area_2_content(title='Locations')
    ui.on('exlbl', exlbl)
    def exlbp(event):
        nonlocal gui
        app.storage.user['tool_query'], *_ = event.args
        gui.select_empty_area2_tab()
        gui.load_area_2_content(title='Characters')
    ui.on('exlbp', exlbp)
    def exlbt(event):
        nonlocal gui
        app.storage.user['tool_query'], *_ = event.args
        gui.select_empty_area2_tab()
        gui.load_area_2_content(title='Topics')
    ui.on('exlbt', exlbt)
    def bibleDict(event):
        nonlocal gui
        app.storage.user['tool_query'], *_ = event.args
        gui.select_empty_area2_tab()
        gui.load_area_2_content(title='Dictionaries')
    ui.on('bibleDict', bibleDict)
    def encyclopedia(event):
        nonlocal gui
        app.storage.user['favorite_encyclopedia'], app.storage.user['tool_query'] = event.args
        gui.select_empty_area2_tab()
        gui.load_area_2_content(title='Encyclopedias')
    ui.on('encyclopedia', encyclopedia)

    # --- CONFIGURATION ---
    DB_FILE = os.path.join(BIBLEMATEWEB_DATA, 'indexes2.sqlite')

    # Define your tables, their display titles, and their icons
    TABLE_CONFIG = {
        "Bible People": {
            "table": "exlbp", 
            "icon": "groups"        # Icon for people/groups
        },
        "Bible Locations": {
            "table": "exlbl", 
            "icon": "place"         # Icon for maps/locations
        },
        "Bible Topics": {
            "table": "exlbt", 
            "icon": "category"      # Icon for topics/categories
        },
        "Bible Dictionaries": {
            "table": "dictionaries", 
            "icon": "loyalty"     # Icon for definitions/books
        },
        "Bible Encyclopedia": {
            "table": "encyclopedia", 
            "icon": "diamond" # Icon for deep knowledge/library
        }
    }

    def fetch_data(table_name: str, book_id: int, chapter: int, verse: int):
        """
        Connects to SQLite and retrieves information for the specific verse.
        """
        try:
            with apsw.Connection(DB_FILE) as connection:
                cursor = connection.cursor()
                query = f"SELECT Information FROM {table_name} WHERE Book=? AND Chapter=? AND Verse=?"
                cursor.execute(query, (book_id, chapter, verse))
                rows = cursor.fetchall()
                return "\n\n".join([row[0].replace("</td><td>", "&nbsp;&nbsp;</td><td>") for row in rows]) if rows else None
        except Exception as e:
            return f"Error querying database: {str(e)}"

    # --- UI LAYOUT ---
    with ui.column().classes('w-full max-w-3xl mx-auto p-4 gap-6'):

        # Bible Selection menu
        bible_selector = BibleSelector(version_options=["KJV"])
        def additional_items():
            nonlocal gui, bible_selector, area
            def change_indexes(selection):
                if area == 1:
                    app.storage.user['tool_book_text'], app.storage.user['bible_book_number'], app.storage.user['bible_chapter_number'], app.storage.user['bible_verse_number'] = selection
                    gui.load_area_1_content(title="Indexes")
                else:
                    app.storage.user['tool_book_text'], app.storage.user['tool_book_number'], app.storage.user['tool_chapter_number'], app.storage.user['tool_verse_number'] = selection
                    gui.load_area_2_content(title="Indexes", sync=False)
            ui.button(get_translation('Go'), on_click=lambda: change_indexes(bible_selector.get_selection()))
        bible_selector.create_ui("KJV", b, c, v, additional_items=additional_items, show_versions=False)

        # Results Container
        results_container = ui.column().classes('w-full gap-4')

    def search_action(book_id, chapter, verse):
        # Clear UI and show loading spinner
        results_container.clear()
        
        with results_container:
            
            # Run IO-bound DB operations
            results = {}
            for title, config in TABLE_CONFIG.items():
                data = fetch_data(config['table'], book_id, chapter, verse)
                results[title] = data
            
            #parser = BibleVerseParser(False, language=app.storage.user['ui_language'])
            #ui.label(f"Resources for {parser.bcvToVerseReference(book_id, chapter, verse)}") \
            #    .classes('text-xl font-semibold text-secondary mb-2')

            # --- RENDER EXPANSIONS ---
            found_any = False
            
            for title, config in TABLE_CONFIG.items():
                content = results.get(title)
                
                # Check if this should be open by default
                is_open = (title == "Bible Topics")
                
                # Create the Expansion with specific icon
                with ui.expansion(title, icon=config['icon'], value=is_open) \
                        .classes('w-full border rounded-lg shadow-sm') \
                        .props('header-class="font-bold text-lg text-secondary"'):
                    
                    if content:
                        # convert links, e.g. <ref onclick="bcv(3,19,26)">
                        content = content.replace("<table><tr>", "<tr>")
                        content = content.replace("<tr>", "<table><tr>")
                        content = content.replace("</tr></table>", "</tr>")
                        content = content.replace("</tr>", "</tr></table>")
                        content = re.sub(r'''<ref onclick="(searchEncyc|searchDict)\('.*?'\)">(.*?)</ref>''', r"\2", content)
                        content = re.sub(r'''(onclick|ondblclick)="(cr|bcv|website|exlbl|exlbp|exlbt|bibleDict|encyclopedia)\((.*?)\)"''', r'''\1="emitEvent('\2', [\3]); return false;"''', content)
                        content = re.sub(r"""(onclick|ondblclick)='(cr|bcv|website|exlbl|exlbp|exlbt|bibleDict|encyclopedia)\((.*?)\)'""", r"""\1='emitEvent("\2", [\3]); return false;'""", content)
                        ui.html(f'<div class="content-text">{content}</div>', sanitize=False).classes('p-4')
                        found_any = True
                    else:
                        ui.label('No entries found for this category.') \
                            .classes('p-4 italic')

            def open_collection(collection, tool, number):
                nonlocal gui
                app.storage.user['tool_query'] = f"{tool}.{number}"
                gui.select_empty_area2_tab()
                gui.load_area_2_content(title=collection)

            for i in ("PROMISES_INDEXES", "PARALLEL_INDEXES"):

                # Create the Expansion with specific icon
                with ui.expansion("Bible Promises" if i == "PROMISES_INDEXES" else "Bible Parallels", icon="redeem" if i == "PROMISES_INDEXES" else "link", value=is_open) \
                        .classes('w-full border rounded-lg shadow-sm') \
                        .props('header-class="font-bold text-lg text-secondary"'):

                    DB_FILE2 = os.path.join(BIBLEMATEWEB_DATA, "collections3.sqlite")
                    sql_query = "SELECT Tool, Number, Topic FROM " + i + " WHERE Passages LIKE ?"
                    with apsw.Connection(DB_FILE2) as connn:
                        cursor = connn.cursor()
                        cursor.execute(sql_query, (f"%({book_id}, {chapter}, {verse})%",))
                        fetches = cursor.fetchall()
                    
                    if fetches:
                        for tool, number, topic in fetches:
                            ui.chip(
                                topic,
                                icon='book',
                                color='primary',
                                on_click=partial(open_collection, "Promises" if i == "PROMISES_INDEXES" else "Parallels", tool, number),
                            ).props('text-color=white').classes('cursor-pointer font-bold shadow-sm')
                    else:
                        ui.label('No entries found for this category.') \
                            .classes('p-4 italic')

            if not found_any:
                ui.notify('No data found in any index for this verse.', type='info')
        
    search_action(b,c,v)



from biblemateweb import BIBLEMATEWEB_DATA, config, load_vectors_from_db, get_translation
from nicegui import ui, app, run
from agentmake.utils.rag import get_embeddings, cosine_similarity_matrix
import numpy as np
import re, apsw, os, json, traceback
from biblemateweb.data.cr_books import cr_books
from biblemateweb.fx.shared import get_image_data_uri

def fetch_bible_encyclopedias_entry(path, sql_table):
    db = os.path.join(BIBLEMATEWEB_DATA, "data", "encyclopedia.data")
    with apsw.Connection(db) as connn:
        cursor = connn.cursor()
        sql_query = f"SELECT content FROM {sql_table} WHERE path=? limit 1"
        cursor.execute(sql_query, (path,))
        fetch = cursor.fetchone()
        content = fetch[0] if fetch else ""
    return content

async def fetch_bible_encyclopedias_matches_async(query, sql_table):
    n = ui.notification("Loading ...", timeout=None, spinner=True)
    db_file = os.path.join(BIBLEMATEWEB_DATA, "vectors", "encyclopedia.db")
    embedding_model = "paraphrase-multilingual"
    path = ""
    options = []

    try:
        # 1. Exact Match Check (Fast, run on main thread)
        exact_match_found = False
        with apsw.Connection(db_file) as connection:
            cursor = connection.cursor()
            cursor.execute(f"SELECT * FROM {sql_table} WHERE entry = ?;", (query,))
            rows = cursor.fetchall()
            
            if len(rows) == 1:
                path = rows[0][0]
                exact_match_found = True
            elif len(rows) > 1:
                options = [f"[{row[0]}] {row[1]}" for row in rows]
                exact_match_found = True

        # 2. Similarity Search (Offloaded to CPU bound thread)
        if not exact_match_found:
            # A. Get Query Vector (IO Bound)
            query_vector = await run.io_bound(get_embeddings, [query], embedding_model)
            query_vector = query_vector[0]

            # B. Load Vectors (The Fix: Run in separate process)
            # We pass the 'sql_table' argument dynamically here
            entries, document_matrix = await run.cpu_bound(load_vectors_from_db, db_file, sql_table)

            if not entries:
                return path, []

            # C. Compute Similarity (CPU Bound)
            similarities = await run.cpu_bound(cosine_similarity_matrix, query_vector, document_matrix)
            
            # D. Sort & Select Top Matches
            top_indices = np.argsort(similarities)[::-1][:app.storage.user["top_similar_entries"]]
            options = [entries[i] for i in top_indices]

    except Exception as ex:
        print("Error during database operation:", ex)
        traceback.print_exc()
        n.message = f'Error: {str(ex)}'
        n.type = 'negative'
        return
    finally:
        n.dismiss()
        
    return path, options

def fetch_bible_encyclopedias_matches(query, sql_table):
    db_file = os.path.join(BIBLEMATEWEB_DATA, "vectors", "encyclopedia.db")
    embedding_model="paraphrase-multilingual"
    path = ""
    options = []
    try:
        with apsw.Connection(db_file) as connection:
            # search for exact match first
            cursor = connection.cursor()
            cursor.execute(f"SELECT * FROM {sql_table} WHERE entry = ?;", (query,))
            rows = cursor.fetchall()
            if not rows: # perform similarity search if no an exact match
                # convert query to vector
                query_vector = get_embeddings([query], embedding_model)[0]
                # fetch all entries
                cursor.execute(f"SELECT path, entry, entry_vector FROM {sql_table}")
                all_rows = [(f"[{path}] {entry}", entry_vector) for path, entry, entry_vector in cursor.fetchall()]
                if not all_rows:
                    return []
                # build a matrix
                entries, entry_vectors = zip(*[(row[0], np.array(json.loads(row[1]))) for row in all_rows if row[0] and row[1]])
                document_matrix = np.vstack(entry_vectors)
                # perform a similarity search
                similarities = cosine_similarity_matrix(query_vector, document_matrix)
                top_indices = np.argsort(similarities)[::-1][:app.storage.user["top_similar_entries"]]
                # return top matches
                options = [entries[i] for i in top_indices]
            elif len(rows) == 1: # single exact match
                path = rows[0][0]
            else:
                options = [f"[{row[0]}] {row[1]}" for row in rows]
    except Exception as ex:
        print("Error during database operation:", ex)
        traceback.print_exc()
        ui.notify('Error during database operation!', type='negative')
        return
    return path, options

def fetch_all_encyclopedias(sql_table):
    db = os.path.join(BIBLEMATEWEB_DATA, "vectors", "encyclopedia.db")
    with apsw.Connection(db) as connn:
        cursor = connn.cursor()
        sql_query = f"SELECT entry FROM {sql_table}"
        cursor.execute(sql_query)
        all_entries = [i[0] for i in cursor.fetchall()]
    return list(set([i for i in all_entries if i]))

def search_bible_encyclopedias(gui=None, q='', **_):

    last_entry = q
    scope_select = None

    def cr(event):
        nonlocal gui
        b, c, v, *_ = event.args
        b = cr_books.get(b, b)
        gui.change_area_1_bible_chapter(None, b, c, v)

    def bcv(event):
        nonlocal gui
        b, c, v, *_ = event.args
        gui.change_area_1_bible_chapter(None, b, c, v)
    
    def website(event):
        url, *_ = event.args
        ui.navigate.to(url, new_tab=True)

    ui.on('bcv', bcv)
    ui.on('cr', cr)
    ui.on('website', website)

    sql_table = app.storage.user.get('favorite_encyclopedia', 'ISB')

    # --- Fuzzy Match Dialog ---
    with ui.dialog() as dialog, ui.card().classes('w-full max-w-md'):
        ui.label("Bible Encyclopedia ...").classes('text-xl font-bold text-secondary mb-4')
        ui.label("We couldn't find an exact match. Please select one of these topics:").classes('text-secondary mb-4')
        
        # This container will hold the radio selection dynamically
        selection_container = ui.column().classes('w-full')
        
        with ui.row().classes('w-full justify-end mt-4'):
            ui.button('Cancel', on_click=dialog.close).props('flat color=grey')

    # ----------------------------------------------------------
    # Core: Fetch and Display
    # ----------------------------------------------------------

    async def show_entry(path, keep=True):
        nonlocal content_container, gui, dialog, input_field, sql_table, last_entry

        n = ui.notification(get_translation('Loading...'), timeout=None, spinner=True)
        content = await run.io_bound(fetch_bible_encyclopedias_entry, path, sql_table)
        n.dismiss()

        # update tab records
        if content and keep:
            gui.update_active_area2_tab_records(q=path)

        # Clear existing rows first
        content_container.clear()

        with content_container:
            # convert links, e.g. <ref onclick="bcv(3,19,26)">
            content = re.sub(r'''(onclick|ondblclick)="(cr|bcv|website)\((.*?)\)"''', r'''\1="emitEvent('\2', [\3]); return false;"''', content)
            content = re.sub(r"""(onclick|ondblclick)='(cr|bcv|website)\((.*?)\)'""", r"""\1='emitEvent("\2", [\3]); return false;'""", content)
            # remove map
            content = content.replace('<div id="map" style="width:100%;height:500px"></div>', "")
            content = re.sub(r'<script.*?>.*?</script>', '', content, flags=re.DOTALL)
            # convert colors for dark mode, e.g. <font color="brown">
            if app.storage.user['dark_mode']:
                content = content.replace('color="brown">', 'color="pink">')
                content = content.replace('color="navy">', 'color="lightskyblue">')
                content = content.replace('<table bgcolor="#BFBFBF"', '<table bgcolor="#424242"')
                content = content.replace('<td bgcolor="#FFFFFF">', '<td bgcolor="#212121">')
                content = content.replace('<tr bgcolor="#FFFFFF">', '<tr bgcolor="#212121">')
                content = content.replace('<tr bgcolor="#DFDFDF">', '<tr bgcolor="#303030">')
            # convert images to data URI
            def replace_img(match):
                img_module = match.group(1)
                img_src = match.group(2)
                img_src = f"{img_module}_{img_src}"
                data_uri = get_image_data_uri(img_module, img_src)
                if data_uri:
                    return f'<img style="display: inline-block;" src="{data_uri}"/>'
                else:
                    return match.group(0)  # return original if not found
            content = re.sub(r'<img src="getImage.php\?resource=([A-Z]+?)&id=(.+?)"/>', replace_img, content)
            # display
            ui.html(f'<div class="content-text">{content}</div>', sanitize=False)

            with ui.row().classes('w-full justify-center q-my-md'):
                ui.button('Show All Verses', icon='auto_stories', on_click=lambda: gui.show_all_verses(path)) \
                    .props('size=lg rounded color=primary')

        # Clear input so user can start typing to filter immediately
        last_entry = path
        input_field.value = ""

    def handle_up_arrow():
        nonlocal last_entry, input_field
        if not input_field.value.strip():
            input_field.value = last_entry

    async def handle_enter(e, keep=True):
        nonlocal sql_table, dialog, input_field, last_entry, selection_container

        query = input_field.value.strip()

        modules = "|".join(list(config.encyclopedias.keys()))
        if not query:
            return
        elif re.search(f"^(ISBE|{modules})[0-9]+?$", query):
            await show_entry(query, keep=keep)
            return

        last_entry = query
        input_field.disable()

        try:
            path, options = await fetch_bible_encyclopedias_matches_async(query, sql_table)
        except Exception as e:
            # Handle errors (e.g., network failure)
            ui.notify(f'Error: {e}', type='negative')

        finally:
            # ALWAYS re-enable the input, even if an error occurred above
            input_field.enable()
            # Optional: Refocus the cursor so the user can type the next query immediately
            input_field.run_method('focus')

        if options:
            options = list(set(options))
            def handle_selection(selected_option):
                nonlocal dialog
                if selected_option:
                    dialog.close()
                    path, _ = selected_option.split(" ", 1)
                    ui.timer(0, lambda: show_entry(path[1:-1], keep=keep), once=True)

            selection_container.clear()
            with selection_container:
                # We use a radio button for selection
                radio = ui.radio(options).classes('w-full').props('color=primary')
                ui.button('Show Content', on_click=lambda: handle_selection(radio.value)) \
                    .classes('w-full mt-4 bg-blue-500 text-white shadow-md')    
            dialog.open()
        else:
            await show_entry(path, keep=keep)

    # ==============================================================================
    # 3. UI LAYOUT
    # ==============================================================================
    client_encyclopedias = list(config.encyclopedias.keys())
    if q and ":::" in q:
        additional_options, q = q.split(":::", 1)
        if additional_options.strip() in client_encyclopedias:
            app.storage.user['favorite_encyclopedia'] = additional_options.strip()

    with ui.row().classes('w-full max-w-3xl mx-auto m-0 py-0 px-4 items-center'):
        scope_select = ui.select(
            options=client_encyclopedias,
            value=app.storage.user.get('favorite_encyclopedia', 'ISB'),
            with_input=True
        ).classes('w-22').props('dense')

        input_field = ui.input(
            autocomplete=[],
            placeholder=f'{get_translation("Search")} {config.encyclopedias[sql_table]} ...'
        ).classes('flex-grow text-lg') \
        .props('outlined dense clearable autofocus enterkeyhint="search"')

        input_field.on('keydown.enter.prevent', handle_enter)
        input_field.on('keydown.up', handle_up_arrow)
        with input_field.add_slot('append'):
            ui.icon('history') \
                .on('click', handle_up_arrow) \
                .classes('text-sm cursor-pointer text-secondary').tooltip('Restore last entry')

        async def get_all_entries(sql_table):
            all_entries = await run.io_bound(fetch_all_encyclopedias, sql_table)
            input_field.set_autocomplete(all_entries)
        n = ui.notification(get_translation('Loading...'), timeout=None, spinner=True)
        ui.timer(0, get_all_entries, once=True)
        n.dismiss()

        async def handle_scope_change(e):
            nonlocal sql_table
            sql_table = e.value
            app.storage.user['favorite_encyclopedia'] = sql_table
            n = ui.notification(get_translation('Loading...'), timeout=None, spinner=True)
            all_entries = await run.io_bound(fetch_all_encyclopedias, sql_table)
            n.dismiss()
            input_field.set_autocomplete(all_entries)
            input_field.props(f'placeholder="Search {config.encyclopedias[sql_table]} ..."')
        scope_select.on_value_change(handle_scope_change)

    # --- Main Content Area ---
    with ui.column().classes('w-full items-center'):
        # Define the container HERE within the layout structure
        content_container = ui.column().classes('w-full transition-all !gap-1')

    if q:
        input_field.value = q
        ui.timer(0, lambda: handle_enter(None, keep=False), once=True)
    else:
        input_field.run_method('focus')


from agentmake.plugins.uba.lib.BibleBooks import BibleBooks
from biblemateweb import BIBLEMATEWEB_DATA, get_translation
from agentmake.plugins.uba.lib.BibleParser import BibleVerseParser
from biblemateweb.data.cr_books import cr_books
from nicegui import ui, app
import re, apsw, os

def fetch_tske(b, c, v):
    db = os.path.join(BIBLEMATEWEB_DATA, "cross-reference.sqlite")
    with apsw.Connection(db) as connn:
        sql_query = "SELECT Information FROM TSKe WHERE Book=? AND Chapter=? AND Verse=? limit 1"
        cursor = connn.cursor()
        cursor.execute(sql_query, (b, c, v))
        fetch = cursor.fetchone()
    return fetch[0] if fetch else ""

def treasury(gui=None, b=1, c=1, v=1, q='', **_):

    last_entry = q
    SQL_QUERY = "PRAGMA case_sensitive_like = false; SELECT Book, Chapter, Verse, Scripture FROM Verses WHERE (Scripture REGEXP ?) ORDER BY Book, Chapter, Verse"

    # --- Data: 66 Bible Books & ID Mapping ---
    BIBLE_BOOKS = [BibleBooks.abbrev["eng"][str(i)][0] for i in range(1,67)]

    def cr(event):
        nonlocal gui
        b, c, v, *_ = event.args
        b = cr_books.get(b, b)
        gui.change_area_1_bible_chapter(None, b, c, v)

    ui.on('cr', cr)

    # ----------------------------------------------------------
    # Core: Fetch and Display
    # ----------------------------------------------------------
    def handle_up_arrow():
        nonlocal last_entry, input_field
        if not input_field.value.strip():
            input_field.value = last_entry

    def handle_enter(e, keep=True):
        nonlocal gui, SQL_QUERY, input_field, last_entry, content_container

        query = input_field.value.strip()
        if not query:
            return
        last_entry = query
        parser = BibleVerseParser(False)
        refs = parser.extractAllReferences(query)
        if not refs:
            ui.notify('No verses found!', type='negative')
            return
        # update tab records
        if keep:
            gui.update_active_area2_tab_records(q=query)
        # Clear existing rows first
        content_container.clear()

        input_field.disable()

        try:

            results = []
            for ref in refs:
                results.append(f"<h2>{parser.bcvToVerseReference(*ref)}</h2>")
                for ref2 in parser.extractExhaustiveReferences([ref]):
                    b, c, v = ref2
                    query = ""
                    content = fetch_tske(b, c, v)
                    if content:
                        results.append(content)
                    else:
                        continue
            content = "<hr>".join(results) if results else ""
            if not content: return None

        except Exception as e:
            # Handle errors (e.g., network failure)
            ui.notify(f'Error: {e}', type='negative')

        finally:
            # ALWAYS re-enable the input, even if an error occurred above
            input_field.enable()
            # Optional: Refocus the cursor so the user can type the next query immediately
            input_field.run_method('focus')

        with content_container:
            # convert links, e.g. <ref onclick="bcv(3,19,26)">
            content = re.sub(r'''(onclick|ondblclick)="(bdbid|lex|cr|bcv|website)\((.*?)\)"''', r'''\1="emitEvent('\2', [\3]); return false;"''', content)
            content = re.sub(r"""(onclick|ondblclick)='(bdbid|lex|cr|bcv|website)\((.*?)\)'""", r"""\1='emitEvent("\2", [\3]); return false;'""", content)
            # convert colors for dark mode, e.g. <font color="brown">
            if app.storage.user['dark_mode']:
                content = re.sub("<red>|</red>", "", content)
            ui.html(f'<div class="content-text">{content}</div>', sanitize=False)
                    

    # ==============================================================================
    # 3. UI LAYOUT
    # ==============================================================================
    with ui.row().classes('w-full max-w-3xl mx-auto m-0 py-0 px-4 items-center'):
        input_field = ui.input(
            autocomplete=BIBLE_BOOKS,
            placeholder=get_translation("Enter bible verse reference(s) here...")
        ).classes('flex-grow text-lg') \
        .props('outlined dense clearable autofocus enterkeyhint="search"')

        input_field.on('keydown.enter.prevent', handle_enter)
        #input_field.on('update:model-value', filter_verses)
        input_field.on('keydown.up', handle_up_arrow)
        with input_field.add_slot('append'):
            ui.icon('history') \
                .on('click', handle_up_arrow) \
                .classes('text-sm cursor-pointer text-secondary').tooltip('Restore last entry')

    # --- Main Content Area ---
    with ui.column().classes('w-full items-center'):
        # Define the container HERE within the layout structure
        content_container = ui.column().classes('w-full transition-all !gap-1')

    if q:
        input_field.value = q
    else:
        parser = BibleVerseParser(False, language=app.storage.user['ui_language'])
        input_field.value = parser.bcvToVerseReference(b,c,v)
    handle_enter(None, keep=False)
    input_field.run_method('focus')


from nicegui import ui, app
from biblemateweb import get_translation


# --- LOGIC: Bible Selection Dialog ---
class BibleSelectionDialog:
    def __init__(self, parent, book_names, verses):
        with ui.dialog() as self.dialog, ui.card().classes('w-[90vw] max-w-screen-lg h-[90vh] flex flex-col p-4') as self.content_area:
            pass
        self.selection = {'book_id': None, 'chapter': None, 'verse': None}
        self.parent = parent
        self.book_names = book_names
        self.verses = verses
        self.set_current_verse()
    
    def set_current_verse(self):
        bible_tab = self.parent.get_active_area1_tab()
        if bible_tab in app.storage.user:
            args = app.storage.user.get(bible_tab)
            self.current_verse = (args.get("b"), args.get("c"), args.get("v"))
        else:
            self.current_verse = (1, 1, 1)

    def open(self):
        self.set_current_verse()
        self.selection = {'book_id': None, 'chapter': None, 'verse': None}
        self.render_books()
        self.dialog.open()

    def close(self):
        self.dialog.close()

    def render_header(self, title):
        # shrink-0 ensures the header doesn't get compressed by the scroll area
        with ui.row().classes('w-full justify-between items-center mb-2 shrink-0'):
            ui.label(title).classes('text-xl font-bold')
            ui.button(icon='close', on_click=self.close).props('flat round dense')
        ui.separator().classes('mb-2 shrink-0')

    def render_books(self):
        # Clear existing content in the reused card
        self.content_area.clear()
        
        with self.content_area:
            self.render_header(get_translation("Select Book"))
            
            # Use a generic container with native scrolling for better mobile compatibility
            # Use flex-1 and min-h-0 to ensure proper scrolling behavior inside flex container
            with ui.element('div').classes('w-full flex-1 overflow-y-auto min-h-0'):
                # 3-column grid for better space utilization
                with ui.row().classes('w-full gap-2'):
                    for book_id in range(1, 67):
                        name = self.book_names.get(book_id, f"Book {book_id}")
                        if book_id == self.current_verse[0]:
                            ui.button(name, on_click=lambda id=book_id: self.select_book(id)) \
                                .props('outline align=center text-color=secondary').classes('w-20')
                        else:
                            ui.button(name, on_click=lambda id=book_id: self.select_book(id)) \
                                .props('align=center').classes('w-20')                            

    def select_book(self, book_id):
        self.selection['book_id'] = book_id
        self.render_chapters()

    def render_chapters(self):
        self.content_area.clear()
        book_name = self.book_names.get(self.selection['book_id'])
        
        with self.content_area:
            self.render_header(f"{book_name}: {get_translation('Select Chapter')}")
            
            # Use flex-1 and min-h-0 to ensure proper scrolling behavior inside flex container
            with ui.element('div').classes('w-full flex-1 overflow-y-auto min-h-0'):
                # Reduced columns from 10 to 6 for better mobile layout
                with ui.row().classes('w-full gap-2 p-2'):
                    chapters = self.verses.get(self.selection['book_id'], {})
                    for chap_num in chapters.keys():
                        if chap_num == self.current_verse[1]:
                            ui.button(str(chap_num), on_click=lambda c=chap_num: self.select_chapter(c)) \
                                .classes('w-15').props('outline align=center text-color=secondary')
                        else:
                            ui.button(str(chap_num), on_click=lambda c=chap_num: self.select_chapter(c)) \
                                .classes('w-15')

            ui.button(get_translation("Back"), on_click=self.render_books).props('outline color=grey').classes('w-full mt-2 shrink-0')

    def select_chapter(self, chapter):
        self.selection['chapter'] = chapter
        self.render_verses()

    def render_verses(self):
        self.content_area.clear()
        book_name = self.book_names.get(self.selection['book_id'])
        
        with self.content_area:
            self.render_header(f"{book_name} {self.selection['chapter']}: {get_translation('Select Verse')}")
            
            # Get max verses for this chapter
            max_verses = self.verses[self.selection['book_id']][self.selection['chapter']]
            
            # Use flex-1 and min-h-0 to ensure proper scrolling behavior inside flex container
            with ui.element('div').classes('w-full flex-1 overflow-y-auto min-h-0'):
                # Reduced columns from 10 to 6 for better mobile layout
                with ui.row().classes('w-full gap-2 p-2'): 
                    for v in range(1, max_verses + 1):
                        if v == self.current_verse[2]:
                            ui.button(str(v), on_click=lambda verse=v: self.finish_selection(verse)) \
                                .classes('w-15').props('outline align=center text-color=secondary')
                        else:
                            ui.button(str(v), on_click=lambda verse=v: self.finish_selection(verse)) \
                                .classes('w-15')

            ui.button(get_translation("Back"), on_click=self.render_chapters).props('outline color=grey').classes('w-full mt-2 shrink-0')

    def finish_selection(self, verse):
        self.selection['verse'] = verse
        self.close()
        
        # Trigger the final python function
        #book_name = self.book_names.get(self.selection['book_id'])
        #msg = f"Selected: {book_name} {self.selection['chapter']}:{self.selection['verse']}"
        #ui.notify(msg, type='positive', icon='check_circle', close_button=True, position='center')
        self.parent.change_area_1_bible_chapter(book=self.selection['book_id'], chapter=self.selection['chapter'], verse=self.selection['verse'])


from nicegui import ui, app
import asyncio, os, re, glob
from biblemateweb import get_translation, BIBLEMATEWEB_DATA
from biblemateweb.fx.bible import BibleSelector
from agentmake.plugins.uba.lib.BibleBooks import BibleBooks
from agentmake import readTextFile


class BiblePodcastPlayer:
    def __init__(self, text_list: list, audio_list: list, start_verse=1, title="Bible Podcast", next_book_callback=None):
        self.title = title
        self.text_list = text_list
        self.audio_list = audio_list
        self.current_verse = None
        self.is_playing = False
        self.audio_element = None
        self.verse_buttons = {}
        self.start_verse = start_verse
        self.next_book_callback = next_book_callback
        
    def create_ui(self):

        with ui.card().classes('w-full max-w-4xl mx-auto mt-8 p-6'):
            # Title
            self.reference = ui.label(self.title).classes('text-3xl font-bold mb-6 text-center')
            
            # Audio player and controls container
            with ui.row().classes('w-full items-center justify-between mb-6 gap-4'):
                # Audio player
                self.audio_element = ui.audio('').classes('flex-grow')
                self.audio_element.on('ended', self.on_audio_ended)
                
                # Loop toggle
                with ui.row().classes('items-center gap-2'):
                    ui.label(get_translation("Loop")).classes('text-sm font-medium')
                    self.loop_toggle = ui.switch().bind_value(app.storage.user, 'loop_podcast')
            
            ui.separator().classes('mb-4')
            # Verse list
            with ui.column().classes('w-full gap-2'):
                with ui.list().props('bordered separator').classes('w-full'):
                    for i in self.text_list:
                        if found := re.search("^.*?_([0-9]+?)_", i):
                            num_str = found.group(1)
                            verse_num = int(num_str[1:]) if len(num_str) > 1 and num_str.startswith("0") else int(num_str)
                            verse_text = re.sub("^.*?_([0-9]+?)_", "", i)
                            verse_text = verse_text[:-4].replace("_", " ")
                        else:
                            continue
                        with ui.item().classes(f'w-full hover:bg-gray-{500 if app.storage.user["dark_mode"] else 50}'):
                            with ui.item_section().props('avatar'):
                                # Audio control button
                                btn = ui.button(icon='volume_off', 
                                            on_click=lambda v=verse_num: self.toggle_verse(v))
                                btn.classes('flat round color=primary')
                                self.verse_buttons[verse_num] = btn
                            with ui.item_section():
                                verse_text = f"<vid>{self.title} {verse_num}</vid> {verse_text}" if verse_num else f"<vid>{self.title}</vid> {verse_text}"
                                ui.html(verse_text, sanitize=False).classes('text-base')

    def toggle_verse(self, verse_num):
        if self.current_verse == verse_num and self.is_playing:
            # Stop current verse
            self.stop_playing()
        else:
            # Play selected verse
            self.play_verse(verse_num)
    
    def play_verse(self, verse_num):
        # Stop current playing verse
        if self.current_verse is not None:
            self.verse_buttons[self.current_verse].props('icon=volume_off')
        
        # Update state
        self.current_verse = verse_num
        self.is_playing = True
        self.reference.text = f"{self.title} {verse_num}" if verse_num else self.title
        
        # Update UI
        self.verse_buttons[verse_num].props('icon=volume_up')
        
        # Load and play audio
        audio_file = self.audio_list[verse_num]
        self.audio_element.set_source(audio_file)
        self.audio_element.run_method('play')
    
    def stop_playing(self):
        if self.audio_element:
            self.audio_element.run_method('pause')
        
        if self.current_verse is not None:
            self.verse_buttons[self.current_verse].props('icon=volume_off')
        
        self.is_playing = False
        self.current_verse = None
    
    def on_audio_ended(self):
        # Current verse finished
        if self.current_verse is not None:
            self.verse_buttons[self.current_verse].props('icon=volume_off')
        
        # Determine next verse
        if self.current_verse is not None:
            next_verse = self.current_verse + 1
            
            # Check if we've reached the end
            if next_verse >= len(self.text_list):
                if app.storage.user.get("loop_podcast"):
                    # Loop back to verse 1
                    self.play_verse(1)
                else:
                    # Stop playing
                    self.is_playing = False
                    self.current_verse = None
                    if self.next_book_callback is not None:
                        self.next_book_callback()
            else:
                # Play next verse
                self.play_verse(next_verse)
    
    async def auto_start(self):
        # Wait a bit for the page to fully load
        await asyncio.sleep(0.5)
        # Start playing from the specified verse
        self.play_verse(self.start_verse)

def bibles_podcast(gui=None, bt=None, b=1, c=1, v=1, area=2, **_):
    if not bt:
        bt = gui.get_area_1_bible_text()

    def next_book(selected_b):
        nonlocal gui
        new_b = selected_b+1 if selected_b < 66 else 1
        app.storage.user["tool_book_number"], app.storage.user["tool_chapter_number"], app.storage.user["tool_verse_number"] = new_b, 1, 1
        gui.load_area_2_content(title="Podcast", sync=False)

    # version
    version = "KJV"
    # Bible Selection menu
    bible_selector = BibleSelector(version_options=["KJV"])
    def additional_items():
        nonlocal gui, bible_selector, area
        def change_audio_chapter(selection):
            if area == 1:
                app.storage.user['tool_book_text'], app.storage.user['bible_book_number'], app.storage.user['bible_chapter_number'], app.storage.user['bible_verse_number'] = selection
                gui.load_area_1_content(title="Podcast")
            else:
                app.storage.user['tool_book_text'], app.storage.user['tool_book_number'], app.storage.user['tool_chapter_number'], app.storage.user['tool_verse_number'] = selection
                gui.load_area_2_content(title="Podcast", sync=False)
        ui.button(get_translation('Go'), on_click=lambda: change_audio_chapter(bible_selector.get_selection()))
    bible_selector.create_ui(version, b, c, v, additional_items=additional_items, show_versions=False, show_verses=False)
    # audio folder
    bible_audio_dir = os.path.join(BIBLEMATEWEB_DATA, "podcast")
    book_dir = ""
    for i in os.listdir(bible_audio_dir):
        if num_str_found := re.search("^([0-9]+?)_", i):
            num_str = num_str_found.group(1)
            num = int(num_str[1:]) if len(num_str) > 1 and num_str.startswith("0") else int(num_str)
            if num == b:
                book_dir = os.path.join(bible_audio_dir, i)
                break
    if not book_dir: return
    playlist_file = glob.glob(os.path.join(book_dir, "*.m3u"))
    if not playlist_file: return
    playlist_file = playlist_file[0]
    # Text file list
    text_list = [i for i in readTextFile(playlist_file).split("\n") if i]
    # Audio file list
    audio_list = [os.path.join(book_dir, i) for i in text_list]
    # Start the player
    player = BiblePodcastPlayer(text_list=text_list, audio_list=audio_list, start_verse=c, title=BibleBooks.abbrev[app.storage.user["ui_language"]][str(b)][-1], next_book_callback=lambda: next_book(b))
    player.create_ui()
    # Auto-start playing after page loads
    ui.timer(0.5, player.auto_start, once=True)



from nicegui import ui, app
from biblemateweb.data.bible_events import bible_events
from biblemateweb import get_translation

def bible_chronology(gui=None, **_):
    #ui.page_title('Bible Chronology')

    # Function to be triggered when the link is clicked
    def open_verse_reference(reference: str):
        """Updates the output label with information about the clicked reference."""
        nonlocal gui
        app.storage.user['tool_query'] = reference
        gui.select_empty_area2_tab()
        gui.load_area_2_content(title='Verses')

    # Function to create the clickable link component
    def create_internal_link(reference: str):
        # 1. Create the ui.link component (it looks like a link)
        link_component = ui.link(
            text=reference,
            target='#' # Set target to '#' or omit it since we are not opening a URL
        ).classes('text-secondary hover:text-primary underline')
        # 2. Attach the Python function to the 'click' event
        # We use a lambda to pass the specific 'reference' string argument to the function
        link_component.on('click', lambda: open_verse_reference(reference))
        
        return link_component

    with ui.card().classes('w-full'):
        ui.label('📜 Bible Chronology').classes('text-2xl font-bold mb-4')
        with ui.timeline(side='right'):
            for item in bible_events:
                # Create the interactive link
                ref_link = create_internal_link(item['reference'])
                
                with ui.timeline_entry(
                    title=item['event'],
                    subtitle=item['year'],
                    icon='menu_book'
                ):
                    with ui.row():
                        ref_link


from nicegui import ui
#from nicegui import app
from functools import partial
from agentmake.plugins.uba.lib.BibleParser import BibleVerseParser

def parousia_zh(gui=None, **_):
    parser = BibleVerseParser(False)
    def open_verse_reference(reference: str):
        """Updates the output label with information about the clicked reference."""
        nonlocal gui, parser
        #app.storage.user['tool_query'] = reference
        #gui.select_empty_area2_tab()
        #gui.load_area_2_content(title='Verses')
        if refs := parser.extractAllReferences(reference):
            b, c, v, *_ = refs[0]
            gui.change_area_1_bible_chapter(book=b, chapter=c, verse=v)
        else:
            ui.notify("No verse reference found!")

    def render_verse_chips(verses):
        """Helper to render a list of verses as dense chips."""
        with ui.row().classes('gap-1 flex-wrap'):
            for verse in verses:
                # clickable=True makes them interactive if you add on_click later
                ui.chip(
                    verse,
                    icon='book',
                    on_click=partial(open_verse_reference, verse),
                ).props('dense outline square text-color=white icon=menu_book').classes('bg-white text-xs')

    # Page Container
    with ui.column().classes('w-full max-w-4xl mx-auto p-4 gap-6'):
        
        # --- Header ---
        with ui.column().classes('w-full items-center text-center mb-6'):
            ui.label('圍繞降臨 (Parousia) 的事件順序').classes('text-3xl font-bold text-primary')
            ui.label('觀點：前千禧年派 (Premillennialism) & 災後被提 (Post-tribulation)').classes('text-lg font-medium')
            ui.label('主要經文架構：哥林多前書 15:23-27, 保羅書信, 啟示錄 20 章').classes('text-gray-500')

        # --- Expansion 1: The Return of Jesus ---
        with ui.expansion('耶穌的再來，降臨 (The Return of Jesus, the Parousia)', icon='auto_awesome') \
                .classes('w-full bg-blue-50 rounded-lg border border-blue-200'):
            with ui.column().classes('p-4 w-full'):
                
                ui.label('● 耶穌的禱告').classes('font-bold text-blue-900')
                render_verse_chips(['太 6:10', '路 11:2'])
                
                ui.separator().classes('mb-2')
                
                ui.label('● 新創造').classes('font-bold text-blue-900')
                render_verse_chips(['賽 65:17', '賽 66:22'])
                
                ui.separator().classes('mb-2')

                ui.label('● 再來').classes('font-bold text-blue-900')
                render_verse_chips([
                    '約 14:3', '林前 15:23', '帖前 2:19', '帖前 3:13', 
                    '帖前 4:15', '帖前 5:23', '帖後 2:1', '帖後 2:8', 
                    '雅 5:7', '彼後 1:16', '彼後 3:4', '彼後 3:12', '約一 2:28'
                ])

        # --- Expansion 2: Witnesses surrounding the Parousia ---
        with ui.expansion('圍繞降臨 (Parousia) 事件的見證', icon='verified') \
                .classes('w-full bg-purple-50 rounded-lg border border-purple-200'):
            with ui.column().classes('p-4 w-full'):
                
                ui.label('● 一千年').classes('font-bold text-purple-900')
                render_verse_chips(['啟 20:6'])

                ui.separator().classes('mb-2')

                ui.label('● 舊約中神在地上的統治').classes('font-bold text-purple-900')
                render_verse_chips([
                    '詩 72:8-14', '賽 2:2-4', '但 2:28-45', 
                    '彌 4:1-3', '亞 14:5-17', '賽 11:1-10'
                ])

                ui.separator().classes('mb-2')

                ui.label('● 與人類共同掌權').classes('font-bold text-purple-900')
                render_verse_chips(['創 1:26-28', '詩 8:4-6', '來 2:6-10'])

                ui.separator().classes('mb-2')

                ui.label('● 舊約中作為列國之光的祭司國度').classes('font-bold text-purple-900')
                render_verse_chips(['出 19:6', '賽 42:6', '賽 49:6'])

                ui.separator().classes('mb-2')

                ui.label('● 君尊的祭司').classes('font-bold text-purple-900')
                render_verse_chips([
                    '彼前 2:9', '啟 1:6', '啟 5:10', '出 19:5-6', 
                    '創 1:26-28', '太 19:28', '路 22:30', 
                    '提後 2:12', '啟 20', '啟 5:10', '啟 20:4', '啟 22:5'
                ])

                ui.separator().classes('mb-2')

                ui.label('● 亞當被造是為了統治').classes('font-bold text-purple-900')
                render_verse_chips(['西 1:16', '詩 2', '詩 110'])

                ui.separator().classes('mb-2')

                ui.label('● 移交給聖徒').classes('font-bold text-purple-900')
                render_verse_chips(['但 7:27'])

        # --- Timeline ---
        with ui.timeline(side='right').classes('w-full'):

            # =================================================
            # STAGE 1: Jesus' Resurrection
            # =================================================
            with ui.timeline_entry(
                title='第一階段：耶穌的復活',
                subtitle='初熟的果子',
                icon='church',
                color='green-6'
            ):
                with ui.card().classes('bg-green-50 w-full mt-2 border-l-4 border-green-500'):
                    ui.markdown('**(1) 耶穌的復活**')
                    # Implied from the header context "1 Cor 15:23"
                    render_verse_chips(['林前 15:23'])

            # =================================================
            # CONTEXT: Post-tribulation (Before Parousia)
            # =================================================
            with ui.timeline_entry(
                title='前奏：災後 (Post-tribulation)',
                subtitle='大災難在降臨之前',
                icon='warning',
                color='red-6'
            ):
                with ui.card().classes('bg-red-50 w-full mt-2 border-l-4 border-red-500'):
                    ui.label('大災難 (The Great Tribulation)').classes('font-bold text-red-900')
                    render_verse_chips([
                        '約 16:33', '帖前 5:2-12', '帖後 2:9-10', 
                        '啟 13:10', '啟 14:12', '馬可 13:19-25', 
                        '太 24:21-29', '西 1:24', '啟 1:9'
                    ])
                    
                    ui.separator().classes('my-3 bg-red-200')
                    
                    ui.label('第二次再來之前 (Pre-Second Coming Events)').classes('font-bold text-red-900')
                    render_verse_chips([
                        '帖後 2:1-12', '帖後 2:3', '帖前 4:17', 
                        '約一 4:3', '約二 1:7', '啟 13:1 - 啟 17:18', 
                        '啟 7:1-8, 9, 14'
                    ])

            # =================================================
            # STAGE 2: The Parousia
            # =================================================
            with ui.timeline_entry(
                title='第二階段：降臨 (Parousia)',
                subtitle='信徒復活與彌賽亞統治',
                icon='flight_land',
                color='blue-7'
            ):
                with ui.card().classes('bg-blue-50 w-full mt-2 border-l-4 border-blue-500'):
                    ui.markdown('**i. 突然的降臨**')
                    render_verse_chips(['腓 3:20', '帖前 5:1-4'])

                    ui.separator().classes('my-3 bg-blue-200')

                    ui.markdown('**ii. 已死的信徒復活**')
                    render_verse_chips([
                        '林前 15:51-52', '腓 3:21', '帖前 4:13-17', 
                        '帖前 5:1-4', '啟 20:4-6'
                    ])

                    ui.separator().classes('my-3 bg-blue-200')

                    ui.markdown('**iii. 彌賽亞的過渡統治期 (Messianic Interregnum)**')
                    ui.label('末了權勢的爭戰').classes('text-sm text-gray-700 mb-1')
                    render_verse_chips([
                        '羅 16:20', '林前 15:22-24', '林前 6:3', '啟 20:4'
                    ])

            # =================================================
            # STAGE 3: The End
            # =================================================
            with ui.timeline_entry(
                title='第三階段：末期',
                subtitle='最終的結局',
                icon='public',
                color='purple-6'
            ):
                with ui.card().classes('bg-purple-50 w-full mt-2 border-l-4 border-purple-500'):
                    ui.markdown('**i. 所有人的普遍復活**')
                    render_verse_chips([
                        '約 5:29', '啟 20:4-6', '啟 20:13'
                    ])

                    ui.separator().classes('my-3 bg-purple-200')

                    ui.markdown('**ii. 最後的審判**')
                    render_verse_chips([
                        '羅 14:10', '林後 5:10'
                    ])

                    ui.separator().classes('my-3 bg-purple-200')

                    ui.markdown('**iii. 萬物/被造界的轉變與更新**')
                    render_verse_chips(['羅 8:19-20'])


from agentmake.plugins.uba.lib.BibleBooks import BibleBooks
from biblemateweb import BIBLEMATEWEB_DATA, get_translation
from biblemateweb.fx.bible import get_bible_content
from agentmake.plugins.uba.lib.BibleParser import BibleVerseParser
from functools import partial
from nicegui import ui, app, run
import re, apsw, os

def fetch_xrefs(b, c, v):
    db = os.path.join(BIBLEMATEWEB_DATA, "cross-reference.sqlite")
    with apsw.Connection(db) as connn:
        sql_query = "SELECT Information FROM ScrollMapper WHERE Book=? AND Chapter=? AND Verse=? limit 1"
        cursor = connn.cursor()
        cursor.execute(sql_query, (b, c, v))
        query = cursor.fetchone()
        return query

def xrefs(gui=None, b=1, c=1, v=1, q='', **_):

    last_entry = q
    SQL_QUERY = "PRAGMA case_sensitive_like = false; SELECT Book, Chapter, Verse, Scripture FROM Verses WHERE (Scripture REGEXP ?) ORDER BY Book, Chapter, Verse"

    # --- Data: 66 Bible Books & ID Mapping ---
    BIBLE_BOOKS = [BibleBooks.abbrev["eng"][str(i)][0] for i in range(1,67)]


    # ----------------------------------------------------------
    # Helper: Filter Logic
    # ----------------------------------------------------------
    def filter_verses(e=None):
        """
        Filters visibility based on input.
        Iterates over default_slot.children to find rows.
        """
        total_matches = 0
        # Robustly determine the search text
        text = ""
        if e is not None and hasattr(e, 'value'):
            text = e.value 
        else:
            text = input_field.value 
            
        search_term = text.lower() if text else ""
        
        # Iterate over the actual children of the container
        for row in verses_container.default_slot.children:
            # Skip elements that aren't our verse rows (if any)
            if not hasattr(row, 'verse_data'):
                continue

            # Explicitly show all if search is empty
            if not search_term:
                row.set_visibility(True)
                continue

            data = row.verse_data
            ref_text = data['ref'].lower()
            clean_content = re.sub('<[^<]+?>', '', data['content']).lower()

            is_match = (search_term in ref_text) or (search_term in clean_content)
            row.set_visibility(is_match)
            if is_match:
                total_matches += 1
        if total_matches:
            ui.notify(f"{total_matches} {'match' if total_matches == 1 else 'matches'} found!")

    # ----------------------------------------------------------
    # Helper: Remove Verse
    # ----------------------------------------------------------
    def remove_verse_row(row_element, reference):
        try:
            verses_container.remove(row_element)
            ui.notify(f'Removed: {reference}', type='warning', position='top')
        except Exception as e:
            print(f"Error removing row: {e}")

    # ----------------------------------------------------------
    # Helper: Open Chapter
    # ----------------------------------------------------------
    def open_chapter_next_area2_tab(bible, b, c, v):
        gui.select_next_area2_tab()
        gui.change_area_2_bible_chapter(bible, b, c, v, sync=False)

    def open_chapter_empty_area2_tab(bible, b, c, v):
        gui.select_empty_area2_tab()
        gui.change_area_2_bible_chapter(bible, b, c, v, sync=False)

    # ----------------------------------------------------------
    # Core: Fetch and Display
    # ----------------------------------------------------------
    def handle_up_arrow():
        nonlocal last_entry, input_field
        if not input_field.value.strip():
            input_field.value = last_entry

    async def handle_enter(e, keep=True):
        nonlocal SQL_QUERY, input_field, verses_container, last_entry
        query = input_field.value.strip()
        if not query:
            return
        last_entry = query
        parser = BibleVerseParser(False)
        refs = parser.extractAllReferences(query)
        if not refs:
            ui.notify('No verses found!', type='negative')
            return
        # update tab records
        if keep:
            gui.update_active_area2_tab_records(q=query)
        # Clear existing rows first
        verses_container.clear()

        input_field.disable()

        try:

            for ref in refs:
                for ref2 in parser.extractExhaustiveReferences([ref]):
                    b, c, v = ref2
                    query = ""
                    n = ui.notification(get_translation('Loading...'), timeout=None, spinner=True)
                    query = await run.io_bound(fetch_xrefs, b, c, v)
                    n.dismiss()
                    if query:
                        query = query[0]
                    else:
                        #ui.notify('No verses found!', type='negative')
                        continue
                    
                    # Prepend the original verse reference to the query for context
                    query = parser.bcvToVerseReference(b, c, v) + "; " + query
                    
                    if not query:
                        #ui.notify('Display cleared', type='positive', position='top')
                        continue

                    parser = BibleVerseParser(False, language=app.storage.user['ui_language'])
                    verses = get_bible_content(query, bible=gui.get_area_1_bible_text(), sql_query=SQL_QUERY, parser=parser)

                    if not verses:
                        #ui.notify('No verses found!', type='negative')
                        continue

                    with verses_container:
                        
                        ui.label("Cross-References - " + verses[0]['ref']).classes('text-2xl font-serif text-secondary')
                        
                        for v in verses:
                            # Row setup
                            with ui.column().classes('w-full shadow-sm rounded-lg items-start no-wrap border border-gray-200 !gap-0') as row:
                                
                                row.verse_data = v # Store data for filter function

                                # --- Chip (Clickable & Removable) ---
                                with ui.element('div').classes('flex-none pt-1'): 
                                    with ui.chip(
                                        v['ref'], 
                                        removable=True, 
                                        icon='book',
                                        #on_click=partial(ui.notify, f'Clicked {v['ref']}'),
                                    ).props('text-color=white').classes('cursor-pointer font-bold shadow-sm') as chip:
                                        with ui.menu():
                                            ui.menu_item('Open in Bible Area', on_click=partial(gui.change_area_1_bible_chapter, v['bible'], v['b'], v['c'], v['v']))
                                            ui.menu_item('Open Here', on_click=partial(gui.change_area_2_bible_chapter, v['bible'], v['b'], v['c'], v['v'], sync=False))
                                            ui.menu_item('Open in Next Tab', on_click=partial(open_chapter_next_area2_tab, v['bible'], v['b'], v['c'], v['v']))
                                            ui.menu_item('Open in New Tab', on_click=partial(open_chapter_empty_area2_tab, v['bible'], v['b'], v['c'], v['v']))
                                    chip.on('remove', lambda _, r=row, ref=v['ref']: remove_verse_row(r, ref))

                                # --- Content ---
                                ui.html(v['content'], sanitize=False).classes('grow min-w-0 leading-relaxed pl-2 text-base break-words')

                    # Clear input so user can start typing to filter immediately
                    input_field.value = ""
                    input_field.props(f'''placeholder="{get_translation('Type to filter')} {len(verses)} {get_translation('results')}..."''')
                    ui.notify(f"{len(verses)} {get_translation('result') if not verses or len(verses) == 1 else get_translation('results')}")

        except Exception as e:
            # Handle errors (e.g., network failure)
            ui.notify(f'Error: {e}', type='negative')

        finally:
            # ALWAYS re-enable the input, even if an error occurred above
            input_field.enable()
            # Optional: Refocus the cursor so the user can type the next query immediately
            input_field.run_method('focus')

    # ==============================================================================
    # 3. UI LAYOUT
    # ==============================================================================
    with ui.row().classes('w-full max-w-3xl mx-auto m-0 py-0 px-4 items-center'):
        input_field = ui.input(
            autocomplete=BIBLE_BOOKS,
            placeholder=get_translation("Enter bible verse reference(s) here...")
        ).classes('flex-grow text-lg') \
        .props('outlined dense clearable autofocus enterkeyhint="search"')

        input_field.on('keydown.enter.prevent', handle_enter)
        input_field.on('update:model-value', filter_verses)
        input_field.on('keydown.up', handle_up_arrow)
        with input_field.add_slot('append'):
            ui.icon('history') \
                .on('click', handle_up_arrow) \
                .classes('text-sm cursor-pointer text-secondary').tooltip('Restore last entry')

    # --- Main Content Area ---
    with ui.column().classes('w-full items-center'):
        # Define the container HERE within the layout structure
        verses_container = ui.column().classes('w-full transition-all !gap-1')

    if q:
        input_field.value = q
    else:
        parser = BibleVerseParser(False, language=app.storage.user['ui_language'])
        input_field.value = parser.bcvToVerseReference(b,c,v)
    ui.timer(0, lambda: handle_enter(None, keep=False), once=True)
    input_field.run_method('focus')


from nicegui import ui, app, run
import os, asyncio, apsw
from biblemateweb import BIBLEMATEWEB_DATA, get_translation
from agentmake.plugins.uba.lib.BibleBooks import BibleBooks

def get_db_file(lang="eng"):
    if lang == "tc":
        basefile = "book_analysis_tc.data"
    elif lang == "sc":
        basefile = "book_analysis_sc.data"
    else:
        basefile = "book_analysis.data"
    return os.path.join(BIBLEMATEWEB_DATA, "data", basefile)

def get_contents(b, lang="eng"):
    fetches = None
    contents = []
    with apsw.Connection(get_db_file(lang)) as connn:
        query = "SELECT Content FROM Introduction WHERE Book=? ORDER BY Section"
        cursor = connn.cursor()
        cursor.execute(query, (b,))
        fetches = cursor.fetchall()
    if fetches:
        contents = [fetch[0] for fetch in fetches]
    return contents

def book_analysis(gui=None, b=1, q="", **_):

    # handle book change
    def on_book_change(event):
        nonlocal gui
        app.storage.user['tool_book_number'] = event.value
        gui.load_area_2_content(title="Analysis", sync=False)

    # handle bcv events
    def bcv(event):
        nonlocal gui
        b, c, v, *_ = event.args
        gui.change_area_1_bible_chapter(None, b, c, v)
    ui.on('bcv', bcv)

    # Summary display
    async def load_analysis(b):
        n = ui.notification("Loading ...", timeout=None, spinner=True)
        await asyncio.sleep(0)
        contents = await run.io_bound(get_contents, b, app.storage.user['ui_language'])

        with ui.row().classes('w-full justify-center'):
            book_options = {i: BibleBooks.abbrev[app.storage.user['ui_language']][str(i)][-1] for i in range(1,67)}
            ui.select(
                options=book_options,
                value=b,
                on_change=on_book_change,
            )

        if contents:
            sections = {
                "Overview": "info",
                "Structural Outline": "account_tree",
                "Logical Flow": "low_priority",
                "Historical Setting": "history_edu",
                "Themes": "style",
                "Keywords": "vpn_key",
                "Theology": "auto_stories",
                "Canonical Placement": "format_list_numbered",
                "Practical Living": "volunteer_activism",
                "Summary": "summarize",
            }
            # display content
            index = 0
            for section, icon in sections.items():
                with ui.expansion(get_translation(section), icon=icon, value=(q.lower() == section.lower())) \
                            .classes('w-full border rounded-lg shadow-sm') \
                            .props('header-class="font-bold text-lg text-secondary"'):
                    ui.html(f'<div class="content-text">{contents[index]}</div>', sanitize=False)
                index += 1
        n.dismiss()

    ui.timer(0, lambda: load_analysis(b), once=True)    


def get_original_css(dark_mode):
    return f"""
<style>
    ntgloss, otgloss, inter {{
        direction: ltr;
        display: inline-block;
        font-size: 1.1rem;
        color: {'#f2ac7e' if dark_mode else '#D35400'};
    }}
    /* links */
    ref {{
        color: {'#f2c522' if dark_mode else 'navy'};
        font-weight: bold;
        cursor: pointer;
    }}

    /* Each verse acts as a container for word blocks */
    verse {{
        display: block;
        margin-bottom: 20px;
        line-height: 1.3;
    }}
    /* The interlinear word block container */
    .int {{
        display: inline-block;
        vertical-align: top;
        text-align: center;
        margin: 0px;
        padding: 4px 8px;
        border-radius: 6px;
        box-shadow: 0 2px 4px rgba(0,0,0,0.05);
        border: 1px solid #eee;
        transition: background-color 0.2s; /* Smooth highlight transition */
    }}
    /* Target all block elements inside .int to tighten their spacing */
    .int > *, .int > ref > * {{
        line-height: 1.1;
        margin-top: 1px;
        margin-bottom: 1px;
    }}
    /* Hover effects for interactive elements (even if JS isn't active here) */
    .int:hover {{
        background-color: {'#333333' if dark_mode else '#f8fbff'};
        border-color: #d6eaf8;
    }}

    /* Transliteration Layers (Phonetic & SBL) - shared with Greek*/
    wsbl, wphono {{
        direction: ltr;
        display: inline-block;
        font-size: 0.8rem;
        color: #7f8c8d;
        font-style: italic;
    }}
    /* Morphology Layer */
    wmorph {{
        direction: ltr;
        display: inline-block;
        font-family: monospace;
        font-size: 0.7rem;
        color: #27ae60;
        cursor: pointer;
    }}
    wsn {{
        direction: ltr;
        display: inline-block;
        font-size: 0.7rem;
        color: {'#ca96e0' if dark_mode else '#8e44ad'};
        cursor: pointer;
    }}
    /* Gloss (Literal Meaning) Layer */
    wgloss {{
        direction: ltr;
        display: inline-block;
        font-size: 0.85rem;
        color: {'#f2ac7e' if dark_mode else '#D35400'};
    }}
    /* Final Translation Layer */
    wtrans {{
        direction: ltr;
        display: block;
        margin-top: 4px;
        padding-top: 3px;
        border-top: 1px solid #f0f0f0;
        font-size: 0.95rem;
        font-weight: bold;
        color: #2980b9;
        min-height: 1.2em; /* Ensures empty translations don't collapse the block */
    }}

    /* UBA css */
    external {{
        font-size: 0.8rem;
    }}
    red, z {{
        color: red;
    }}
    blu {{
        color: blue;
        font-size: 0.8rem;
    }}
    points {{
        color: gray;
        font-weight: bold;
        font-size: 0.8rem;
    }}
    bb {{
        color: brown;
        font-weight: bold;
    }}
    hp {{
        color: brown;
        font-weight: bold;
        font-size: 0.8rem;
    }}
    highlight {{
        font-style: italic;
    }}
    transliteration {{
        color: gray;
    }}
    div.section, div.point {{
        display: block;
        border: 1px solid green;
        border-radius: 5px;
        padding: 2px 5px;
        margin-top: 5px;
        margin-bottom: 5px;
    }}
    div.remarks {{
        display: block;
        border: 1px solid gray;
        border-radius: 5px;
        padding: 2px 5px;
        margin-top: 5px;
        margin-bottom: 5px;
    }}
    div.bhs, div.hebch {{
        direction: rtl;
    }}
    div.info {{
        margin-left: 5%;
        margin-right: 5%;
    }}
    div.menu {{
        margin-left: 2%;
        margin-right: 2%;
    }}
    div.vword {{
        border: 1px solid #F5B041;
        border-radius: 5px;
        padding: 2px 5px;
        margin-top: 3px;
        margin-bottom: 3px;
    }}
    div.translation {{
        border: 1px solid #9B59B6;
        border-radius: 5px;
        padding: 2px 5px;
        margin-top: 3px;
        margin-bottom: 3px;
    }}
    div.ew {{
        margin-left: 5%;
        margin-right: 5%;
        font-size: 1.1rem;
        display: inline-block;
    }}
    div.mr {{
        margin-left: 100;
    }}
    div.nav {{
        margin-left: 5%;
        margin-right: 5%;
    }}
    div.refList {{
        display: inline;
    }}
    /* css for linguistic annotations */
                
    div.bhp, div.bhw, div.w, div.int {{
        display: inline-block;
        text-align: center;
    }}
    div.int {{
        vertical-align: text-top;
    }}
    div.bhc {{
        direction: rtl;
        border: 1px solid #9B59B6;
        border-radius: 5px;
        padding: 2px 5px;
        margin-top: 3px;
        margin-bottom: 3px;
    }}
    div.bhp {{
        border: 1px solid gray;
        border-radius: 5px;
        padding: 2px 5px;
        margin-top: 3px;
        margin-bottom: 3px;
    }}
    div.bhw {{
        border: 1px solid #F5B041;
        border-radius: 5px;
        padding: 2px 5px;
        margin-top: 3px;
        margin-bottom: 3px;
    }}
    clid, morphCode {{
        color: {'#cfada9' if dark_mode else '#641E16'};
        font-weight: bold;
    }}
    mclid {{
        color: {'#cfada9' if dark_mode else '#641E16'};
        font-style: italic;
    }}
    connector {{
        color: #9B59B6;
    }}
    ckind, ctyp, crela, morphContent {{
        color: {'#f5e798' if dark_mode else '#00008B'};
        display: inline-block;
        direction: ltr;
    }}
    ptyp {{
        text-decoration: underline;
        color: gray;
    }}
    pfunction {{
        color: {'#7ce6d4' if dark_mode else '#3498DB'};
        font-size: 0.9rem;
    }}
    det {{
        font-weight: bold;
    }}
    undet {{
        font-style: italic;
    }}
    prela {{
        display: none;
    }}
    hbint, gntint, gloss {{
        direction: ltr;
        vertical-align: super;
        display: inline-block;
        color: {'#f2ac7e' if dark_mode else '#D35400'};
    }}
    gntint, gloss {{
        font-size: 0.9rem;
    }}
    cllevel {{
        color: #9B59B6;
    }}
    clinfo {{
        color: {'#cfada9' if dark_mode else '#641E16'};
        font-weight: bold;
    }}
    subclinfo {{
        color: {'#cfada9' if dark_mode else '#641E16'};
        font-weight: bold;
        font-size: 0.8rem;
    }}
    funcinfo {{
        color: {'#7ce6d4' if dark_mode else '#3498DB'};
        font-size: 0.9rem;
    }}
    wordid {{
        text-decoration: underline;
        font-size: 0.8rem;
        color: gray;
    }}
    cit, clt, cst, cbhs, cbsb, cleb {{
        display: block;
    }}
    clt, cbsb {{
        color: {'#7ce6d4' if dark_mode else '#3498DB'};
    }}
    cst, cleb {{
        color: {'#f5e798' if dark_mode else '#00008B'};
    }}
    cbsb, cleb {{
        direction: ltr;
    }}
    div.cltrans {{
        margin-left: 10px;
        margin-right: 10px;
        display: block;
        border: 1px solid {'#f5e798' if dark_mode else '#00008B'};
        border-radius: 5px;
        padding: 2px 5px;
        margin-top: 3px;
        margin-bottom: 3px;
    }}
    div.wrap {{
        display: inline-block;
    }}
    cl {{
        display: table;
    }}
    div.c {{
        vertical-align: top;
        text-align: left;
        border: 1px solid #9B59B6;
        border-radius: 5px;
        padding: 2px 5px;
        margin-top: 3px;
        margin-bottom: 3px;
    }}
    div.p {{
        display: inline-block;
        vertical-align: top;
        text-align: left;
        border: 1px solid gray;
        border-radius: 5px;
        padding: 2px 5px;
        margin-top: 3px;
        margin-bottom: 3px;
    }}
    div.w, div.int {{
        border: 1px solid #F5B041;
        border-radius: 5px;
        padding: 2px 5px;
        margin-top: 3px;
        margin-bottom: 3px;
    }}

    /* css for clause segmentation */

    div.bhsa {{
        direction: rtl;
        border-right: 5px solid #F5B041;
        margin-left: 5%;
        margin-right: 5%;
        padding: 5px 10px 5px 10px;
    }}
    div.e {{
        border-left: 5px solid #F5B041;
        margin-left: 5%;
        margin-right: 5%;
        padding: 5px 10px 5px 10px;
    }}
</style>
"""


from nicegui import ui, app

def ai_partner(gui=None, **_):
    with ui.column().classes('w-full items-center justify-center p-8 gap-6 min-h-screen'):
        
        # 1. Visual Indicator
        ui.icon('handshake', size='64px').classes('text-primary mb-4')

        # 2. Main Title
        ui.label('Partner Mode').classes('text-4xl font-bold text-secondary')
        
        # 3. Informative Status Card
        with ui.card().classes('w-full max-w-lg border-l-2 p-6'):
            ui.markdown('**Status:** Currently available in **CLI Version** only.')
            
            ui.label((
                "Partner Mode transforms BibleMate into an interactive study companion. "
                "Instead of just answering questions, it engages in a two-way dialogue "
                "to help you explore texts, brainstorm ideas, and deepen your understanding."
            )).classes(f'mt-4 text-gray-{"300" if app.storage.user["dark_mode"] else "700"} text-lg leading-relaxed')

        # 4. Feature Highlights
        with ui.row().classes('gap-4 mt-4'):
            features = ["Interactive Dialogue", "Joint Exploration", "Study Companion"]
            for feature in features:
                ui.chip(feature, icon='forum').props('text-color=white')

        # 5. Call to Action / Link
        ui.label('To use Partner Mode, please install the BibleMate CLI:').classes(f'mt-8 text-gray-{"400" if app.storage.user["dark_mode"] else "600"}')
        
        ui.button('View on GitHub', icon='open_in_new') \
            .props('href=https://github.com/eliranwong/biblemate target=_blank')



import apsw
import re, os
from nicegui import ui, app
from biblemateweb import BIBLEMATEWEB_DATA, get_translation


def bible_parallels_menu(gui=None, **_):

    def harmony(event):
        nonlocal gui
        tool, number = event.args
        app.storage.user['tool_query'] = f"{tool}.{number}"
        gui.select_empty_area2_tab()
        gui.load_area_2_content(title='Parallels')
    ui.on('harmony', harmony)

    # --- CONFIGURATION ---
    DB_FILE = os.path.join(BIBLEMATEWEB_DATA, 'books', 'Harmonies_and_Parallels.book')
    with apsw.Connection(DB_FILE) as connn:
        cursor = connn.cursor()
        cursor.execute("SELECT Chapter, Content FROM Reference")
        fetches = cursor.fetchall()

    # --- UI LAYOUT ---
    with ui.column().classes('w-full max-w-3xl mx-auto p-4 gap-6'):

        # Results Container
        with ui.column().classes('w-full gap-4'):

            for chapter, content in fetches:
                # Create the Expansion with specific icon
                with ui.expansion(chapter, icon='auto_stories', value=False) \
                        .classes('w-full border rounded-lg shadow-sm') \
                        .props('header-class="font-bold text-lg text-secondary"'):
                    
                    # convert links, e.g. <ref onclick="bcv(3,19,26)">
                    content = re.sub(r'''(onclick|ondblclick)="(cr|bcv|harmony)\((.*?)\)"''', r'''\1="emitEvent('\2', [\3]); return false;"''', content)
                    content = re.sub(r"""(onclick|ondblclick)='(cr|bcv|harmony)\((.*?)\)'""", r"""\1='emitEvent("\2", [\3]); return false;'""", content)
                    ui.html(f'<div class="content-text">{content}</div>', sanitize=False).classes('p-4')
